# -*- coding: utf-8 -*-
"""point polarizability structure class for torchgdm
"""
import warnings
import copy

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm.struct.base_classes import StructBase
from torchgdm.struct.point import StructEffPola3D
from torchgdm.struct.surface_2d import StructDiscretized2D
from torchgdm.tools import interp
from torchgdm.tools.geometry import rotation_x, rotation_y, rotation_z
from torchgdm.tools.geometry import test_structure_distances
from torchgdm.tools.geometry import get_enclosing_sphere_radius
from torchgdm.tools.misc import ptp


# --- base class volume discretized structure container - 3D
class StructEffPola2D(StructEffPola3D):
    """class for 2D line polarizability structure"""

    __name__ = "effective line polarizability (2D) structure class"

    def __init__(
        self,
        positions: torch.Tensor,
        alpha_dicts: list,
        radiative_correction: bool = True,
        device: torch.device = None,
        environment=None,
        shift_z_to_r0: bool = True,
    ):
        """2D line polarizability class

        The main information is provided in the `alpha_dicts`, which is a list of dicts with the full effective polarizability definitions. Each dict defines one structure and must contain following:
            - 'wavelengths': wavelengths at which the polarizabilities are calculated
            - at least one of: ['alpha_pE', 'alpha_mH', 'alpha_mE', 'alpha_pH']:
                polarizability tensors of shape [len(wavelengths), 3, 3]
            optional keys:
            - 'full_geometry': the original volume discretization of the represented geometry
            - 'r0': the origin of the effective polarizabilities with respect to optional 'full_geometry'
            - 'enclosing_radius': enclosing radius of original structure

        Args:
            positions (torch.Tensor): polarizability positions (3D, but all y values must be zero)
            alpha_dicts (list): list of polarizability model dictionaries
            radiative_correction (bool, optional): Whether to add raditative correction for cross section calculations. Defaults to True.
            device (torch.device, optional): Defaults to "cpu".
            environment (_type_, optional): 3D environment class. Defaults to None.
            shift_z_to_r0 (bool, optional): If True, if a position z-value is zero, each polarizability model's z position will be shifted to the height of the effective dipole development center. Defaults to True.

        Raises:
            ValueError: _description_
        """
        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        # expand positions, put single scatterer in list
        _positions = torch.as_tensor(positions, dtype=DTYPE_FLOAT, device=self.device)
        if len(_positions.shape) == 1:
            assert len(_positions) == 3
            _positions = _positions.unsqueeze(0)

        if torch.count_nonzero(_positions[..., 1]) > 0:
            warnings.warn("2D structure. Remove all positions with y!=0.")
            _positions = _positions[_positions[..., 1] != 0]
            if len(_positions) == 0:
                raise ValueError("No mesh positions at y=0. Please check geometry.")

        super().__init__(
            positions=positions,
            alpha_dicts=alpha_dicts,
            environment=environment,
            shift_z_to_r0=shift_z_to_r0,
            device=device,
        )
        
        self.n_dim = 2
        
        self.radiative_correction = radiative_correction

    def __repr__(self, verbose=False):
        """description about structure"""
        out_str = ""
        out_str += (
            "------ 2D effective ED / MD line-dipole polarizabilities object -------"
        )
        out_str += "\n" + " nr. of dipole-pairs:    {}".format(
            len(self.get_all_positions())
        )
        out_str += "\n" + " nominal enclosing circle diameters (nm): {}".format(
            [round(float(f), 1) for f in torch.unique(self.step)]
        )
        if len(self.full_geometries) > 0:
            pos = torch.cat(self.full_geometries)
            out_str += "\n" + " original 2D geometry: "
            out_str += "\n" + "  - replacing nr. of meshpoints: {}".format(len(pos))
            bnds = ptp(pos, dim=0)
            out_str += "\n" + "  - size & position:"
            out_str += "\n" + "        X-extension    :    {:.1f} (nm)".format(bnds[0])
            out_str += "\n" + "        Z-extension    :    {:.1f} (nm)".format(bnds[2])
            out_str += "\n" + "  - center of mass :    ({:.1f}, {:.1f}, {:.1f})".format(
                *[float(f) for f in self.get_center_of_mass()]
            )

        return out_str
    
    # - radiative correction for cross section calc. - 2D case
    # inherit from discretized 2D class
    get_radiative_correction_prefactor_p =  StructDiscretized2D.get_radiative_correction_prefactor_p
    get_radiative_correction_prefactor_m =  StructDiscretized2D.get_radiative_correction_prefactor_m

    # --- plotting
    def plot(
        self,
        scale=1.0,
        color="auto",
        linestyle_circle=(0, (2, 2)),
        color_circle="auto",
        color_circle_fill=None,
        alpha=1,
        show_grid=True,
        color_grid="auto",
        alpha_grid=0.25,
        legend=True,
        set_ax_aspect=True,
        reset_color_cycle=True,
        **kwargs
    ):
        """plot the structure of the effective line-polarizability (2D)

        Args:
            scale (float, optional): scaling factor of the grid cells, if shown. Defaults to 1.0.
            color (str, optional): plot color. Defaults to "auto".
            linestyle_circle (tuple, optional): optional line style for enclosing circle. Defaults to (0, (2, 2)).
            color_circle (str, optional): optional alternative color for enclosing circle. Defaults to "auto".
            color_circle_fill (_type_, optional): optional alternative fill color for enclosing circle. Defaults to None.
            alpha (int, optional): optional transparency. Defaults to 1.
            show_grid (bool, optional): whether to show mesh grid (if available in structure). Defaults to True.
            color_grid (str, optional): optional alternative color for the mesh grid. Defaults to "auto".
            alpha_grid (float, optional): optional alternative transparency for the mesh grid. Defaults to 0.25.
            legend (bool, optional): show legend. Defaults to True.
            set_ax_aspect (bool, optional): automatically set aspect ratio to equal. Defaults to True.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.

        Returns:
            matplotlib axes
        """
        from torchgdm.visu import visu2d

        kwargs["projection"] = "xz"
        im = visu2d.geo2d._plot_structure_eff_pola(
            self,
            scale=scale,
            color=color,
            linestyle_circle=linestyle_circle,
            color_circle=color_circle,
            color_circle_fill=color_circle_fill,
            alpha=alpha,
            show_grid=show_grid,
            color_grid=color_grid,
            alpha_grid=alpha_grid,
            legend=legend,
            set_ax_aspect=set_ax_aspect,
            reset_color_cycle=reset_color_cycle,
            **kwargs
        )
        return im

    def plot_contour(
        self,
        color="auto",
        set_ax_aspect=True,
        alpha=1.0,
        alpha_value=None,
        reset_color_cycle=True,
        **kwargs
    ):
        """plot the contour of the underlying 2D-mesh (2D)

        Args:
            color (str, optional): optional matplotlib compatible color. Defaults to "auto".
            set_ax_aspect (bool, optional): If True, will set aspect of plot to equal. Defaults to True.
            alpha (float, optional): matplotlib transparency value. Defaults to 1.0.
            alpha_value (float, optional): alphashape value. If `None`, try to automatically optimize for best enclosing of structure. Defaults to None.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.

        Returns:
            matplotlib line: matplotlib's `scatter` output
        """
        from torchgdm.visu import visu2d

        kwargs["projection"] = "xz"
        im = visu2d.geo2d.contour(
            self,
            color=color,
            set_ax_aspect=set_ax_aspect,
            alpha=alpha,
            alpha_value=alpha_value,
            reset_color_cycle=reset_color_cycle,
            **kwargs
        )
        return im

    def plot3d(self, **kwargs):
        """plot the point polarizability structure (3D)"""
        from torchgdm.visu import visu3d

        warnings.warn(
            "Visualizing a 2D model in 3D will show only meshpoints in the 'XZ' plane. The 3D plot shows a circumscribing sphere but is in fact a circumscribing cylinder."
        )

        return visu3d.geo3d._plot_structure_eff_3dpola(self, **kwargs)

    # --- geometry operations
    def rotate(
        self,
        alpha: float,
        center: torch.Tensor = torch.as_tensor([0.0, 0.0, 0.0]),
        axis: str = "y",
    ):
        """rotate the structure

        Args:
            alpha (float): rotation angle (in rad)
            center (torch.Tensor, optional): center of rotation axis. Defaults to torch.as_tensor([0.0, 0.0, 0.0]).
            axis (str, optional): rotation axis. Defaults to "y".

        Raises:
            ValueError: only "y" axis supported in 2D

        Returns:
            :class:`StructDiscretized2D`: copy of structure with rotated geometry
        """
        if axis.lower() != "y":
            raise ValueError(
                "Only rotation axis 'y' supported in 2D (infinite axis).".format(axis)
            )

        _struct_rotated = super().rotate(alpha=alpha, center=center, axis=axis)

        return _struct_rotated


# --- Mie sphere
class StructMieCylinderEffPola2D(StructEffPola2D):
    """class for Mie-theory based 2D line polarizability

    Requires external package `treams`
    !!! Does not support automatic differentiation !!!

    Defines a 2D line polarizability representing a cylinder using
    first order (dipolar) Mie coefficients
    """

    __name__ = "Mie-theory cylinder line-dipole polarizability (2D) structure class"

    def __init__(
        self,
        wavelengths: torch.Tensor,
        radii: list,
        materials: list,
        environment=None,
        r0: torch.Tensor = None,
        device: torch.device = None,
        quadrupol_tol=0.15,
    ):
        """2D line polarizability class for a core-shell cylinder (Mie)

        Use Mie theory in dipole approximation (first order) to get an
        effective polarizability model for a core-shell cylinder.
        Requires the `treams` package for Mie coefficient calculation.
        https://github.com/tfp-photonics/treams

        Args:
            wavelengths (torch.Tensor): list of wavelengths to evaluate (nm)
            radii (list): list of the cylinders's core and (multiple) shell radii (in nm).
            materials (list): materials of core and shell(s). A float or int is interpreted as permittivity value.
            environment (environment instance, optional): Homogeneous 3D environment to evaluate Mie theory in. Defaults to None, which uses vacuum.
            r0 (torch.Tensor, optional): polarizability position (x,y,z). If not given, is set to (0, 0, r_cylinder). Defaults to None
            device (torch.device, optional): Defaults to "cpu".
            quadrupol_tol (float, optional): ratio of tolerable residual quadrupole terms relative to the dipole order before warning. Defaults to 0.15.

        Raises:
            ValueError: incorrect parameters
        """
        # prep and imports
        from torchgdm.tools.misc import to_np
        from torchgdm.materials.base_classes import MaterialBase
        from torchgdm.materials import MatConstant
        import numpy as np

        # TODO: Replace with differentiable Mie code
        from torchgdm.tools.mie import mie_ab_cylinder_2d

        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        # tensor conversion
        wavelengths = torch.as_tensor(wavelengths, dtype=DTYPE_FLOAT, device=self.device)
        wavelengths = torch.atleast_1d(wavelengths)
        k0 = 2 * torch.pi / wavelengths

        # mie coefficients
        mie_results = mie_ab_cylinder_2d(
            wavelengths=wavelengths,
            radii=radii,
            materials=materials,
            environment=environment,
            device=self.device,
            m_max=3,
            as_dict=True,
        )
        a_n = mie_results["a_n"]
        b_n = mie_results["b_n"]
        env = mie_results["environment"]
        n_env = mie_results["n_env"]
        r_enclosing = mie_results["r_enclosing"]

        # check if dipole approximation is good
        a_quadrupol_res = a_n[:, 2].abs()
        wls_violation_a = to_np(wavelengths[a_quadrupol_res.to("cpu") > quadrupol_tol])
        if len(wls_violation_a) > 0:
            warnings.warn(
                "Mie series: {} wavelengths with ".format(len(wls_violation_a))
                + "significant residual electric quadrupole contribution: "
                + "{} nm".format([round(r, 1) for r in wls_violation_a])
            )

        b_quadrupol_res = b_n[:, 2].abs()
        wls_violation_b = to_np(wavelengths[b_quadrupol_res.to("cpu") > quadrupol_tol])
        if len(wls_violation_b) > 0:
            warnings.warn(
                "Mie series: {} wavelengths with ".format(len(wls_violation_b))
                + "significant residual magnetic quadrupole contribution: "
                + "{} nm".format([round(r, 1) for r in wls_violation_b])
            )

        # convert to polarizabilities (units of volume)
        # populate 6x6 polarizabilities for all wavelengths
        alpha_6x6 = torch.zeros(
            (len(wavelengths), 6, 6), dtype=DTYPE_COMPLEX, device=self.device
        )
        # TM polarization (s-pol., E parallel axis)
        a_par_pE = 1j / torch.pi * a_n[:, 0] / k0**2
        a_par_mH = 1j * 2 / torch.pi * b_n[:, 0] / k0**2 / n_env**2
        alpha_6x6[:, 1, 1] = a_par_pE
        alpha_6x6[:, 3, 3] = a_par_mH
        alpha_6x6[:, 5, 5] = a_par_mH

        # TE polarization (p-pol., E perp. axis)
        a_perp_pE = 1j * 2 / torch.pi * b_n[:, 1] / k0**2
        a_perp_mH = 1j / torch.pi * a_n[:, 1] / k0**2 / n_env**2
        alpha_6x6[:, 0, 0] = a_perp_pE
        alpha_6x6[:, 2, 2] = a_perp_pE
        alpha_6x6[:, 4, 4] = a_perp_mH

        # set center of mass
        if r0 is None:
            r0 = torch.as_tensor([0.0, 0.0, 0.0], dtype=DTYPE_FLOAT, device=self.device)
        else:
            r0 = torch.as_tensor(r0, dtype=DTYPE_FLOAT, device=self.device)
            r0 = r0.squeeze()
            assert len(r0) == 3
            assert r0[1] == 0  # 2D: must be in XZ plane

        # wrap up in a dictionary compatible with the point dipole structure class
        alpha_dict = dict(
            r0=r0,
            r0_MD=r0,
            r0_ED=r0,
            alpha_6x6=alpha_6x6,
            wavelengths=wavelengths,
            enclosing_radius=r_enclosing,
            k0_spectrum=k0,
        )

        # - point polarizability structure with Mie dipolar response
        super().__init__(
            positions=r0,
            alpha_dicts=[alpha_dict],
            environment=env,
            device=self.device,
        )



# -*- coding: utf-8 -*-
"""
torchgdm tools for geometry data
"""
# %%
from typing import Optional
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device


# --- Cartesian <--> spherical
def transform_spherical_to_xyz(r, teta, phi):
    x = r * torch.sin(teta) * torch.cos(phi)
    y = r * torch.sin(teta) * torch.sin(phi)
    z = r * torch.cos(teta)
    return x, y, z


def transform_xyz_to_spherical(x, y, z):
    r = torch.sqrt(x**2 + y**2 + z**2)
    teta = torch.acos(z / r)
    phi = torch.atan2(y, x)
    return r, teta, phi


# --- coordinate list tools
def coordinate_map_1d(
    lim_r1, n_step_r1=20, r2=0.0, r3=0.0, direction="x", device="cpu"
):
    """Generate Cartesian equidistant coordinates along a line

    Args:
        lim_r1 (tuple of float): run limits coordinate 1
        n_step1 (int, optional): nr of steps coord 1. Defaults to 20.
        r2, r3 (float, optional): fixed value for coord 2 and 3. Defaults to 0.0. Coord order: x,y,z
        direction (str, optional): along which axis. Defaults to 'x'

    Returns:
        dict: contains positions and line elements. keys:
            - "r_probe": torch.Tensor: coordinate list, shape (n_step_r1, 3)
            - "ds": torch.Tensor: list of line elements ("dl") of each position, shape (n_step_r1)
    """
    r1s = torch.linspace(
        lim_r1[0], lim_r1[1], steps=n_step_r1, dtype=DTYPE_FLOAT, device=device
    )

    # point grid
    r2s = r2 * torch.ones_like(r1s)
    r3s = r3 * torch.ones_like(r1s)

    # line elements
    dl = abs(lim_r1[1] - lim_r1[0]) / n_step_r1
    dl = torch.ones_like(r1s) * dl

    # stack x/y/z
    if direction.lower() == "x":
        r_probe = torch.stack([r1s, r2s, r3s], dim=1)
    elif direction.lower() == "y":
        r_probe = torch.stack([r2s, r1s, r3s], dim=1)
    elif direction.lower() == "z":
        r_probe = torch.stack([r2s, r3s, r1s], dim=1)
    else:
        raise ValueError("Unknown direction. Must be one of: x, y, z.")

    dict_surface = dict(r_probe=r_probe, ds=dl)
    return dict_surface


def coordinate_map_1d_circular(
    r=100000.0,
    lim_phi=(0, 2 * torch.pi),
    n_phi=72,
    y0=0,
    device="cpu",
):
    """Generate Cartesian coordinates on a circle in the xz plane

    Equidistant solid angles.
    By default, create coordinates around full circle with radius `r`

    Args:
        r (float, optional): radius of sphere in nm (distance to origin). Defaults to 100000nm (=100 microns). Defaults to 100000.0.
        lim_phi (tuple, optional): minimum and maximum angle in radians, excluding last position (in linear steps from `phimin` to `phimax`). Defaults to (0, 2 * torch.pi).
        n_phi (int, optional): number of angle steps. Defaults to 72.
        y0 (float, optional): y-position of the circle. Defaults to 0.
        device (str, optional): Defaults to "cpu".

    Returns:
        dict: contains positions and line elements. keys:
            - "r_probe": torch.Tensor: coordinate list, shape (n_phi, 3)
            - "ds": torch.Tensor: list of line elements of each position, shape (n_phi)
    """
    to_kws = dict(dtype=DTYPE_FLOAT, device=device)

    # polar angle: exclude endpoint
    phi = (
        torch.ones(int(n_phi), **to_kws)
        * torch.linspace(lim_phi[0], lim_phi[1], int(n_phi) + 1, **to_kws)[:-1]
    )  # .unsqueeze(0)

    # Cartesian points
    x, y, z = transform_spherical_to_xyz(r, torch.ones_like(phi) * torch.pi / 2.0, phi)
    r_probe = torch.stack([x.flatten(), z.flatten(), y.flatten()], axis=1)
    r_probe[:, 1] = y0

    # surface elements
    dphi = abs(lim_phi[1] - lim_phi[0]) / float(n_phi)
    dl = torch.ones_like(phi) * r * dphi

    dict_surface = dict(r_probe=r_probe, ds=dl)
    return dict_surface


def coordinate_map_1d_circular_upper(
    r=100000.0,
    n_phi=72,
    y0=0,
    device="cpu",
):
    """1D Cartesian coordinates in xz plane on the upper half-circle

    teta: pi - 2pi.
    see :func:`coordinate_map_1d_circular` for more detailed documentation.

    Args:
        r (float, optional): radius of circle in nm (distance to origin). Defaults to 100000nm (=100 microns). Defaults to 100000.0.
        n_phi (int, optional): number of angle steps. Defaults to 72.
        y0 (float, optional): y-position of the circle. Defaults to 0.
        device (str, optional): Defaults to "cpu".

    Returns:
        dict: contains positions and line elements. keys:
            - "r_probe": torch.Tensor: coordinate list, shape (n_phi, 3)
            - "ds": torch.Tensor: list of line elements of each position, shape (n_phi)
    """
    return coordinate_map_1d_circular(
        r=r,
        lim_phi=(0, torch.pi),
        n_phi=n_phi,
        y0=y0,
        device=device,
    )


def coordinate_map_1d_circular_lower(
    r=100000.0,
    n_phi=72,
    y0=0,
    device="cpu",
):
    """1D Cartesian coordinates in xz plane on the lower half-circle

    phi: 0 - pi.
    see :func:`coordinate_map_1d_circular` for more detailed documentation.

    Args:
        r (float, optional): radius of circle in nm (distance to origin). Defaults to 100000nm (=100 microns). Defaults to 100000.0.
        n_phi (int, optional): number of angle steps. Defaults to 72.
        y0 (float, optional): y-position of the circle. Defaults to 0.
        device (str, optional): Defaults to "cpu".

    Returns:
        dict: contains positions and line elements. keys:
            - "r_probe": torch.Tensor: coordinate list, shape (n_phi, 3)
            - "ds": torch.Tensor: list of line elements of each position, shape (n_phi)
    """
    return coordinate_map_1d_circular(
        r=r,
        lim_phi=(torch.pi, 2 * torch.pi),
        n_phi=n_phi,
        y0=y0,
        device=device,
    )


def coordinate_map_2d(
    lim_r1,
    lim_r2,
    n_step_r1=20,
    n_step_r2=20,
    r3=0.0,
    projection="xy",
    device="cpu",
):
    """Generate Cartesian equidistant coordinates on rectangular 2d area

    Args:
        lim_r1 (tuple of float): run limits coordinate 1
        lim_r2 (tuple of float): run limits coordinate 2
        n_step1 (int, optional): nr of steps coord 1. Defaults to 20.
        n_step2 (int, optional): nr of steps coord 2. Defaults to 20.
        r3 (float, optional): fixed value for coord 3. Defaults to 0.0.
        projection (str, optional): parallel to which plane. Defaults to 'xy'

    Returns:
        dict: contains positions and surface elements. keys:
            - "r_probe": torch.Tensor: coordinate list, shape (n_step_r1 * n_step_r2, 3)
            - "ds": torch.Tensor: list of surface elements of each position, shape (n_step_r1 * n_step_r2)
    """
    r1s = torch.linspace(
        lim_r1[0], lim_r1[1], steps=n_step_r1, dtype=DTYPE_FLOAT, device=device
    )
    r2s = torch.linspace(
        lim_r2[0], lim_r2[1], steps=n_step_r2, dtype=DTYPE_FLOAT, device=device
    )

    # point grid
    r1f, r2f = torch.meshgrid(r1s, r2s, indexing="xy")
    r1f = r1f.flatten()
    r2f = r2f.flatten()
    r_const = r3 * torch.ones_like(r1f)

    # surface elements
    ds = (abs(lim_r1[1] - lim_r1[0]) / n_step_r1) * (
        abs(lim_r2[1] - lim_r2[0]) / n_step_r2
    )
    ds = torch.ones_like(r1f) * ds

    # stack x/y/z
    if projection.lower() == "xy":
        r_probe = torch.stack([r1f, r2f, r_const], dim=1)
    if projection.lower() == "xz":
        r_probe = torch.stack([r1f, r_const, r2f], dim=1)
    if projection.lower() == "yz":
        r_probe = torch.stack([r_const, r1f, r2f], dim=1)

    dict_surface = dict(r_probe=r_probe, ds=ds)
    return dict_surface


def coordinate_map_2d_square(
    d, n=20, r3=0.0, delta1=0.0, delta2=0.0, projection="xy", device="cpu"
):
    """Generate Cartesian equidistant coordinates on square 2d area

    square map with extents [-d, d] centered at position (delta1, delta2)

    Args:
        d (float): upper and lower limit (useing the negative) for map
        n (int, optional): nr of steps coords 1 and 2. Defaults to 20.
        r3 (float, optional): fixed value for coord 3. Defaults to 0.0.
        delta1, delta2 (float, optional): offsets (shift) for coordinates 1 and 2. Default to 0.0.
        projection (str, optional): parallel to which plane. Defaults to 'xy'
        device (str, optional): which torch device. Defaults to 'cpu'

    Returns:
        dict: contains positions and surface elements. keys:
            - "r_probe": torch.Tensor: coordinate list, shape (n * n, 3)
            - "ds": torch.Tensor: list of surface elements of each position, shape (n * n)
    """
    dict_surface = coordinate_map_2d(
        lim_r1=[-d + delta1, d + delta1],
        lim_r2=[-d + delta2, d + delta2],
        n_step_r1=n,
        n_step_r2=n,
        r3=r3,
        projection=projection,
        device=device,
    )
    return dict_surface


def coordinate_map_2d_spherical(
    r=100000.0,
    lim_teta=(0, torch.pi),
    lim_phi=(0, 2 * torch.pi),
    n_teta=18,
    n_phi=36,
    device="cpu",
):
    """Generate Cartesian coordinate list on a spherical screen with equidistant solid angles

    by default, create coordinates on surface of full sphere with radius `r`

    Args:
        r (float, optional): radius of sphere in nm (distance to origin). Defaults to 100000nm (=100 microns). Defaults to 100000.0.
        lim_teta (tuple, optional): minimum and maximum polar angle in radians (in linear steps from `tetamin` to `tetamax`). Defaults to (0, torch.pi).
        lim_phi (tuple, optional): minimum and maximum azimuth angle in radians, excluding last position (in linear steps from `phimin` to `phimax`). Defaults to (0, 2 * torch.pi).
        n_teta (int, optional): number of polar angle steps. Defaults to 18.
        n_phi (int, optional): number of azimuthal angle steps. Defaults to 36.
        device (str, optional): Defaults to "cpu".

    Returns:
        dict: contains positions and surface elements. keys:
            - "r_probe": torch.Tensor: coordinate list, shape (n_teta * n_phi, 3)
            - "ds": torch.Tensor: list of surface elements of each position, shape (n_teta * n_phi)
    """
    to_kws = dict(dtype=DTYPE_FLOAT, device=device)

    # azimuth
    teta = torch.ones((int(n_teta), int(n_phi)), **to_kws) * torch.linspace(
        lim_teta[0], lim_teta[1], int(n_teta), **to_kws
    ).unsqueeze(1)

    # polar: exclude endpoint
    phi = torch.ones((int(n_teta), int(n_phi)), **to_kws) * torch.linspace(
        lim_phi[0], lim_phi[1], int(n_phi) + 1, **to_kws
    )[:-1].unsqueeze(0)

    # Cartesian points
    x, y, z = transform_spherical_to_xyz(r, teta, phi)
    r_probe = torch.stack([x.flatten(), y.flatten(), z.flatten()], axis=1)

    # surface elements
    dteta = abs(lim_teta[1] - lim_teta[0]) / float(n_teta - 1)
    dphi = abs(lim_phi[1] - lim_phi[0]) / float(n_phi)
    ds = r**2 * torch.sin(teta.flatten()) * dteta * dphi

    dict_surface = dict(r_probe=r_probe, ds=ds)
    return dict_surface


def coordinate_map_2d_spherical_upper(
    r=100000.0,
    n_teta=18,
    n_phi=72,
    device="cpu",
):
    """Generate Cartesian coordinates with equidistant solid angles on the upper half-sphere

    teta: 0 - pi/2, phi: 0 - 2pi.
    see :func:`coordinate_map_2d_spherical` for more detailed documentation.

    Args:
        r (float, optional): radius of sphere in nm (distance to origin). Defaults to 100000nm (=100 microns). Defaults to 100000.0.
        n_teta (int, optional): number of polar angle steps. Defaults to 18.
        n_phi (int, optional): number of azimuthal angle steps. Defaults to 72.
        device (str, optional): Defaults to "cpu".

    Returns:
        dict: contains positions and surface elements. keys:
            - "r_probe": torch.Tensor: coordinate list, shape (n_teta * n_phi, 3)
            - "ds": torch.Tensor: list of surface elements of each position, shape (n_teta * n_phi)
    """
    return coordinate_map_2d_spherical(
        r=r,
        lim_teta=(0, torch.pi / 2),
        lim_phi=(0, 2 * torch.pi),
        n_teta=n_teta,
        n_phi=n_phi,
        device=device,
    )


def coordinate_map_2d_spherical_lower(
    r=100000.0,
    n_teta=18,
    n_phi=72,
    device="cpu",
):
    """Generate Cartesian coordinates with equidistant solid angles on the lower half-sphere

    teta: pi/2 - pi, phi: 0 - 2pi.
    see :func:`coordinate_map_2d_spherical` for more detailed documentation.

    Args:
        r (float, optional): radius of sphere in nm (distance to origin). Defaults to 100000nm (=100 microns). Defaults to 100000.0.
        n_teta (int, optional): number of polar angle steps. Defaults to 18.
        n_phi (int, optional): number of azimuthal angle steps. Defaults to 72.
        device (str, optional): Defaults to "cpu".

    Returns:
        dict: contains positions and surface elements. keys:
            - "r_probe": torch.Tensor: coordinate list, shape (n_teta * n_phi, 3)
            - "ds": torch.Tensor: list of surface elements of each position, shape (n_teta * n_phi)
    """
    return coordinate_map_2d_spherical(
        r=r,
        lim_teta=(torch.pi / 2, torch.pi),
        lim_phi=(0, 2 * torch.pi),
        n_teta=n_teta,
        n_phi=n_phi,
        device=device,
    )


def sample_random_spherical(npoints, device="cpu"):
    """random positions on R=1 sphere"""
    ndim = 3
    vec = torch.rand(npoints, ndim, device=device) - 0.5
    vec /= torch.linalg.norm(vec, axis=1).unsqueeze(1)
    return vec


def sample_random_circular(npoints, projection="xz", device="cpu"):
    """random positions on R=1 circle"""
    if projection.lower() == "xz":
        plane_dim = 1
    elif projection.lower() == "xy":
        plane_dim = 2
    elif projection.lower() == "yz":
        plane_dim = 0
    else:
        raise ValueError

    ndim = 3
    vec = torch.rand(npoints, ndim, device=device) - 0.5
    vec[:, plane_dim] = vec[:, plane_dim] * 0
    vec /= torch.linalg.norm(vec, axis=1).unsqueeze(1)
    return vec


# ----- volume discretization tools
def get_step_from_geometry(positions: torch.Tensor, max_meshpoints: int = 2000):
    """obtain step from a list of coordinates that define a nano-structure

    returns closest distance occuring between two meshpoints in `positions`.
    Uses :func:`torch.cdist`.

    Args:
        positions (torch.Tensor): list of positions to evaluate
        max_meshpoints (int, optional): maximum number of meshpoints to consider for step-calculation (to limit memory requirement). Defaults to 2000.

    Returns:
        float: stepsize between mesh-points
    """
    if len(positions) == 1:
        warnings.warn("Structure consists of a single dipole. Returning `1`.")
        step = 1.0
    elif len(positions) == 0:
        warnings.warn("Empty structure. Returning `1`.")
        step = 1.0
    else:
        if len(positions) > max_meshpoints:
            _geo = positions[:max_meshpoints]
        else:
            _geo = positions
        dists = torch.cdist(_geo, _geo)
        step = torch.sort(torch.flatten(dists))[0][len(_geo)]

    return step


def unique_rows(positions_xy: torch.Tensor, return_inverse: bool = False):
    """delete all duplicates with same x/y coordinates

    Args:
        positions_xy (torch.Tensor): list of (x,y) coordinates with partially redunant (x,y) values.
        return_inverse (bool, optional): return indices of original values which ended up in the unique tensor. Defaults to False.

    Raises:
        ValueError: invalid shape of input positions

    Returns:
        torch.Tensor: tensor containing all unique positions. If `return_inverse` is `True`, return a second torch.Tensor containing the indices in the original tensor of the unique elements
    """
    if positions_xy.shape[1] != 2:
        raise ValueError("Coordinate list must consist of (x,y) 2-tuples.")

    unique_a, idx = torch.unique(positions_xy, dim=0, sorted=True, return_inverse=True)

    if return_inverse:
        # return also indices of unique tensor for each element in the full tensor
        # warning: this is not deterministic, it will return the indices of arbitrarily chosen ones of multiple duplicates
        idx_unique = torch.zeros(
            len(unique_a), dtype=int, device=positions_xy.device
        ).scatter_(0, idx, torch.arange(len(positions_xy), device=positions_xy.device))
        idx_unique.to(unique_a.device)
        return unique_a, idx_unique
    else:
        return unique_a


def get_projection(positions, projection="xy", unique=True, return_inverse=False):
    """get projection of coordinates onto a Cartesian plane

    Args:
        positions (torch.Tensor): _description_
        projection (str, optional): Cartesian plane to project on. One of ["xy", "xz" , "yz"]. Defaults to "xy".
        unique (bool, optional): If True, remove duplicate projected coordinates. Defaults to True.
        return_inverse (bool, optional): return indices of original values which ended up in the unique tensor. Defaults to False.

    Raises:
        ValueError: Invalid projection argument

    Returns:
        torch.Tensor: coordinate list of projected positions. If `return_inverse` is `True`, return a second torch.Tensor containing the indices in the original tensor of the unique elements
    """
    if projection.lower() == "xy":
        twodim_coords = positions[:, [0, 1]]
    elif projection.lower() == "yz":
        twodim_coords = positions[:, [1, 2]]
    elif projection.lower() == "xz":
        twodim_coords = positions[:, [0, 2]]
    else:
        raise ValueError("Invalid projection parameter!")

    if unique and return_inverse:
        pos_proj, idx = unique_rows(twodim_coords, return_inverse=return_inverse)
        return pos_proj, idx
    if unique and not return_inverse:
        pos_proj = unique_rows(twodim_coords)
        return pos_proj
    if not unique:
        return twodim_coords


def get_geometric_crosssection(struct, projection="xy"):
    """return the geometric cross sections ("footprint") projected onto a Cartesian plane

    Args:
        struct (structure instance): instance of structure
        projection (str, optional): Cartesian plane to project on. One of ["xy", "xz" , "yz"]. Defaults to "xy".

    Returns:
        float: geometric cross section in nm^2
    """

    # TODO: calculate this better using alpha shapes?
    step = struct.step
    _pos_proj, idx = get_projection(
        positions=struct.positions, projection=projection, return_inverse=True
    )
    step = step[idx] / struct.mesh_normalization_factor
    cs_geo = torch.sum(step**2)  # * len(pos_proj)
    return cs_geo


def _get_geo_cs_positions_steps(pos, steps, projection="xy"):
    """geometric cross section from list of pos and steps"""
    # TODO: calculate this better using alpha shapes?
    _pos_proj, idx = get_projection(
        positions=pos, projection=projection, return_inverse=True
    )
    steps_visible = steps[idx]
    cs_geo = torch.sum(steps_visible**2)
    return cs_geo


def get_enclosing_sphere_radius(positions: torch.Tensor):
    """get radius of enclosing sphere, that totally contains the full structure

    Args:
        positions (torch.Tensor): (x,y,z) tuples of meshpoints describing the structure

    Returns:
        float: radius of circumscribing sphere
    """
    r0 = torch.mean(positions, axis=0)
    step = get_step_from_geometry(positions)

    geo_center_abs = torch.abs(positions - r0)
    geo_origin_abs = geo_center_abs - torch.min(geo_center_abs, dim=0)[0].unsqueeze(0)
    dist_to_origin = torch.linalg.norm(geo_origin_abs, dim=1)
    dist_to_edge = torch.max(dist_to_origin) + step

    return dist_to_edge


def test_structure_distances(s1, s2, return_counts=True, on_distance_violation="warn"):
    """test if two structures have intersecting elements

    Args:
        s1 (structure-like): torchgdm structure or simulation instance
        s2 (structure-like): torchgdm structure or simulation instance
        return_counts (bool, optional): if True, returns counts of structure 1 and 2 distance-violated positions. Defaults to True.
        on_distance_violation (str, optional): behavior on distance violation. can be "error", "warn" or `None` (do nothing). Defaults to "warn".

    Raises:
        ValueError: on intersection, if `on_distance_violation="error"`

    Returns:
        tuple: return values depend on the `return_counts` kwarg.
            - if `True`: return just 2 numbers: intersecting positions for each struct.
            - if `False`: return 2 lists of positions of distance-violated dipoles in structure 1, respectively 2
    """
    from torchgdm.tools.batch import batched

    @batched(batch_kwarg="pos_step_2", arg_position=1)
    def get_violated_distances(pos_step_1, pos_step_2):
        pos1 = pos_step_1[..., :3]
        pos2 = pos_step_2[..., :3]
        steps1 = pos_step_1[..., 3]
        steps2 = pos_step_2[..., 3]
        dp = torch.norm(pos1.unsqueeze(0) - pos2.unsqueeze(1), dim=-1)
        dp_step1 = torch.where(dp < steps1.unsqueeze(0) / 2 + steps2.unsqueeze(1) / 2)
        dp_step2 = torch.where(dp < steps2.unsqueeze(1) / 2 + steps1.unsqueeze(0) / 2)

        return pos1[torch.unique(dp_step1[1])], pos2[torch.unique(dp_step2[0])]

    pos1 = s1.get_all_positions()
    pos2 = s2.get_all_positions()
    r1 = s1.get_source_validity_radius()
    r2 = s2.get_source_validity_radius()

    pos1_violation, pos2_violation = get_violated_distances(
        pos_step_1=torch.cat([pos1, r1.unsqueeze(1)], dim=1),
        pos_step_2=torch.cat([pos2, r2.unsqueeze(1)], dim=1),
    )

    N_dist1, N_dist2 = len(pos1_violation), len(pos2_violation)

    if on_distance_violation == "error" and (N_dist1 + N_dist2 > 0):
        raise ValueError(
            "Several meshpoints in structures are too close (s1: {}, s2: {})!".format(
                N_dist1, N_dist2
            )
        )
    elif on_distance_violation == "warn" and (N_dist1 + N_dist2 > 0):
        warnings.warn(
            "Several meshpoints in structures are too close (s1: {}, s2: {})!".format(
                N_dist1, N_dist2
            )
        )

    if return_counts:
        return N_dist1, N_dist2
    else:
        return pos1_violation, pos2_violation


def rotation_x(alpha, device="cpu"):
    """matrix for clockwise rotation around x-axis by angle `alpha` (in radian)"""
    alpha = torch.as_tensor(alpha, dtype=DTYPE_FLOAT, device=device)
    s = torch.sin(alpha)
    c = torch.cos(alpha)
    rot_x = torch.as_tensor(
        [[1, 0, 0], [0, c, -s], [0, s, c]],
        dtype=DTYPE_FLOAT,
        device=device,
    )
    return rot_x


def rotation_y(alpha, device="cpu"):
    """matrix for clockwise rotation around y-axis by angle `alpha` (in radian)"""
    alpha = torch.as_tensor(alpha, dtype=DTYPE_FLOAT, device=device)
    s = torch.sin(alpha)
    c = torch.cos(alpha)
    rot_y = torch.as_tensor(
        [[c, 0, s], [0, 1, 0], [-s, 0, c]],
        dtype=DTYPE_FLOAT,
        device=device,
    )
    return rot_y


def rotation_z(alpha, device="cpu"):
    """matrix for clockwise rotation around z-axis by angle alpha (in radian)"""
    alpha = torch.as_tensor(alpha, dtype=DTYPE_FLOAT, device=device)
    s = torch.sin(alpha)
    c = torch.cos(alpha)
    rot_z = torch.as_tensor(
        [[c, -s, 0], [s, c, 0], [0, 0, 1]],
        dtype=DTYPE_FLOAT,
        device=device,
    )
    return rot_z


# %%
if __name__ == "__main__":
    import torchgdm as tg

    # - discretized structure
    struct1 = tg.struct.StructDiscretizedCubic3D(
        tg.struct.volume.cube(l=4), step=25, materials=tg.materials.MatDatabase("Si")
    )
    struct1 = struct1 + [0, 0, 70]
    struct2 = tg.struct.StructDiscretizedCubic3D(
        tg.struct.volume.disc(r=3, h=4),
        step=15,
        materials=tg.materials.MatDatabase("Si"),
    )

    env = tg.env.freespace_3d.EnvHomogeneous3D(
        env_material=tg.materials.MatConstant(eps=1.0)
    )
    # struct2 = struct2.convert_to_effective_polarizability_pair(env, [500])
    # struct2 = struct2.copy([[0, 0, 100], [50, 10, 50]])

    print(test_structure_distances(struct1, struct2, True))

# %%



# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestAutoGrad(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing autograd with discretized simulation...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        # - environment
        mat_env = tg.materials.MatConstant(1)
        self.env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

        # - illumination field(s)
        self.wavelengths = torch.as_tensor([620.0])
        # self.e_inc_list = [
        #     tg.env.freespace_3d.GaussianParaxial(
        #         e0p=0.7, e0s=0.7, inc_angle=torch.pi / 3.0, NA=0.1
        #     ),
        # ]
        self.e_inc_list = [
            tg.env.freespace_3d.PlaneWave(e0p=0.7, e0s=0.7, inc_angle=torch.pi / 3.0),
        ]

        # - structure
        self.step = 20.0
        self.mat = tg.materials.MatDatabase("Ge")
        self.struct = tg.struct.StructDiscretizedCubic3D(
            tg.struct.volume.cuboid(l=2, w=2, h=2), self.step, self.mat
        )

    def test_autograd_wavelength(self):
        for device in self.devices:
            wavelengths = self.wavelengths.clone().to(device=device)
            wavelengths.requires_grad = True

            # setup new material and structure (since wl-grad requires backprop through them)
            mat = tg.materials.MatDatabase("Ge")
            struct = tg.struct.StructDiscretizedCubic3D(
                tg.struct.volume.cuboid(l=2, w=2, h=2), self.step, mat
            )

            # setup sim
            sim = tg.simulation.Simulation(
                structures=[struct],
                environment=self.env,
                illumination_fields=self.e_inc_list,
                wavelengths=wavelengths,
                device=device,
            )

            sim.run(verbose=False, progress_bar=False)
            cs_calc = tg.postproc.crosssect.total(sim, wavelength=wavelengths[0])
            scs_calc = cs_calc["scs"]

            scs_calc.backward()
            dscs_dwl = wavelengths.grad

            # print(scs_calc)
            # print(dscs_dwl)

            # compare to pre-calculated reference
            scs_truth = torch.as_tensor(
                [16.5680466],
                dtype=tg.constants.DTYPE_FLOAT,
                device=device,
            )
            dscs_dwl_truth = torch.as_tensor(
                [-0.1381907],
                dtype=tg.constants.DTYPE_FLOAT,
                device=device,
            )

            torch.testing.assert_close(scs_calc, scs_truth, rtol=1e-3, atol=1e-3)
            torch.testing.assert_close(dscs_dwl, dscs_dwl_truth, rtol=1e-3, atol=1e-3)

            if self.verbose:
                print("  - {}: autograd scs wrt wavelength test passed.".format(device))

    def test_autograd_position(self):
        for device in self.devices:

            pos = torch.as_tensor(
                [5050, 40, 100], dtype=tg.constants.DTYPE_FLOAT, device=device
            )
            pos.requires_grad = True
            self.struct.set_center_of_mass(pos)

            sim = tg.simulation.Simulation(
                structures=[self.struct],
                environment=self.env,
                illumination_fields=self.e_inc_list,
                wavelengths=self.wavelengths,
                device=device,
            )

            sim.run(verbose=False, progress_bar=False)
            cs_calc = tg.postproc.crosssect.total(sim, wavelength=self.wavelengths[0])
            scs_calc = cs_calc["scs"]

            scs_calc.backward()
            dscs_dpos = pos.grad

            # print(scs_calc)
            # print(dscs_dpos)

            # compare to pre-calculated reference
            scs_truth = torch.as_tensor(
                [16.5680466],
                dtype=tg.constants.DTYPE_FLOAT,
                device=device,
            )
            dscs_dpos_truth = torch.as_tensor(
                [8.9406967e-07, 5.9604645e-08, -2.9802322e-07],
                dtype=tg.constants.DTYPE_FLOAT,
                device=device,
            )

            torch.testing.assert_close(scs_calc, scs_truth, rtol=1e-3, atol=1e-3)
            torch.testing.assert_close(dscs_dpos, dscs_dpos_truth, rtol=1e-3, atol=1e-3)

            if self.verbose:
                print(
                    "  - {}: autograd scs wrt structure position test passed.".format(
                        device
                    )
                )


# %%

# %%
if __name__ == "__main__":
    print("testing autograd...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# -*- coding: utf-8 -*-
"""Mie scattering tools using the `treams` toolkit

Mie theory tools (2d and 3d), using the treams t-matrix tool (`pip install treams`)

Beutel, D., Fernandez-Corbaton, I. & Rockstuhl, C. 
**treams – a T-matrix-based scattering code for nanophotonics.**
Computer Physics Communications 297, 109076 (2024)
https://github.com/tfp-photonics/treams

"""
# %%
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm.tools.misc import tqdm, test_illumination_field_config_exists
from torchgdm.tools.misc import to_np


def mie_ab_cylinder_2d(
    wavelengths: torch.Tensor,
    radii: list,
    materials: list,
    environment=None,
    m_max=2,
    device: torch.device = None,
    as_dict=False,
):
    # normal incidence 2D Mie
    from torchgdm.tools.misc import to_np
    from torchgdm.materials.base_classes import MaterialBase
    from torchgdm.materials import MatConstant
    import numpy as np

    # TODO: Replace with differentiable Mie code
    try:
        # ignore import warnings
        with warnings.catch_warnings():
            import treams
    except ModuleNotFoundError:
        print(
            "2D Mie tools require package `treams`. "
            + "Please install via `pip install treams`."
        )
        raise

    if device is None:
        device = get_default_device()
    else:
        device = device

    # environment
    if environment is None:
        from torchgdm.env import EnvHomogeneous2D

        env = EnvHomogeneous2D(device=device)
        warnings.warn("No environment specified. Using a vacuum environment.")
    elif type(environment) in [float, int]:
        from torchgdm.env import EnvHomogeneous2D

        env = EnvHomogeneous2D(env_material=environment, device=device)
        warnings.warn(
            "Environment is a scalar, use as environment permittivity eps={}.".format(
                environment
            )
        )
    else:
        env = environment

    if env.n_dim != 2:
        raise ValueError(
            f"Mie cylinder polarizability requires a 2D environement (is {env.n_dim}D). "
            + "Please use a compatible environment."
        )

    # tensor conversion
    wavelengths = torch.as_tensor(wavelengths, dtype=DTYPE_FLOAT, device=device)
    wavelengths = torch.atleast_1d(wavelengths)
    k0 = 2 * torch.pi / wavelengths
    k0 = torch.as_tensor(k0, device=device)
    kz = 0.0

    # radii to array
    radii = np.atleast_1d(radii)
    r_enclosing = np.max(radii)  # outer radius

    # if single material, put in list
    if not hasattr(materials, "__iter__"):
        materials = [materials]

    # main Mie extraction and setup
    a_n = np.zeros((len(wavelengths), m_max), dtype=np.complex128)
    b_n = np.zeros_like(a_n)
    n_env = np.zeros(len(wavelengths), dtype=np.complex128)
    for i_wl, wl in enumerate(wavelengths):
        # embedding medium and wavevector therein
        eps_env = to_np(env.env_material.get_epsilon(wavelength=wl))[0, 0]
        n_env[i_wl] = eps_env**0.5

        # core and shell materials
        mat_treams = []
        for mat in materials:
            if type(mat) in [float, int, complex]:
                mat = MatConstant(mat)
            eps_mat = to_np(mat.get_epsilon(wavelength=wl))[0, 0]
            mat_treams.append(treams.Material(eps_mat))

        # add environment material last
        mat_treams.append(treams.Material(eps_env))

        for m in range(m_max):
            miecoef = treams.coeffs.mie_cyl(
                kz, m, to_np(k0[i_wl]), radii, *zip(*mat_treams)
            )
            a_n[i_wl, m] = -miecoef[0, 0] - miecoef[0, 1]
            b_n[i_wl, m] = -miecoef[0, 0] + miecoef[0, 1]

    a_n = torch.as_tensor(a_n, device=device, dtype=DTYPE_COMPLEX)
    b_n = torch.as_tensor(b_n, device=device, dtype=DTYPE_COMPLEX)
    n_env = torch.as_tensor(n_env, device=device, dtype=DTYPE_COMPLEX)

    if as_dict:
        return dict(
            a_n=a_n,
            b_n=b_n,
            environment=env,
            n_env=n_env,
            device=device,
            r_enclosing=r_enclosing,
            wavelengths=wavelengths,
        )
    else:
        return a_n, b_n


def mie_ab_sphere_3d(
    wavelengths: torch.Tensor,
    radii: list,
    materials: list,
    environment=None,
    l_max=2,
    device: torch.device = None,
    as_dict=False,
):
    from torchgdm.tools.misc import to_np
    from torchgdm.materials.base_classes import MaterialBase
    from torchgdm.materials import MatConstant
    import numpy as np

    # TODO: Replace with differentiable Mie code
    try:
        # ignore import warnings
        with warnings.catch_warnings():
            import treams
    except ModuleNotFoundError:
        print(
            "3D Mie tools require package `treams`. "
            + "Please install via `pip install treams`."
        )
        raise

    if device is None:
        device = get_default_device()
    else:
        device = device

    # environment
    if environment is None:
        from torchgdm.env import EnvHomogeneous3D

        env = EnvHomogeneous3D(device=device)
        warnings.warn("No environment specified. Using a vacuum environment.")
    elif type(environment) in [float, int]:
        from torchgdm.env import EnvHomogeneous3D

        env = EnvHomogeneous3D(env_material=environment, device=device)
        warnings.warn(
            "Environment is a scalar, use as environment permittivity eps={}.".format(
                environment
            )
        )
    else:
        env = environment

    if env.n_dim != 3:
        raise ValueError(
            f"Mie sphere polarizability requires a 3D environement (is {env.n_dim}D). "
            + "Please use a compatible environment."
        )

    # tensor conversion
    wavelengths = torch.as_tensor(wavelengths, dtype=DTYPE_FLOAT, device=device)
    wavelengths = torch.atleast_1d(wavelengths)
    k0 = 2 * torch.pi / wavelengths
    k0 = torch.as_tensor(k0, device=device)

    # radii to array
    radii = np.atleast_1d(radii)
    r_enclosing = np.max(radii)  # outer radius

    # if single material, put in list
    if not hasattr(materials, "__iter__"):
        materials = [materials]

    # main Mie extraction and setup
    a_n = np.zeros((len(wavelengths), l_max), dtype=np.complex128)
    b_n = np.zeros_like(a_n)
    n_env = np.zeros(len(wavelengths), dtype=np.complex128)
    for i_wl, wl in enumerate(wavelengths):
        # embedding medium and wavevector therein
        eps_env = to_np(env.env_material.get_epsilon(wavelength=wl))[0, 0]
        n_env[i_wl] = eps_env**0.5

        # core and shell materials
        mat_treams = []
        for mat in materials:
            if type(mat) in [float, int, complex]:
                mat = MatConstant(mat)
            eps_mat = to_np(mat.get_epsilon(wavelength=wl))[0, 0]
            mat_treams.append(treams.Material(eps_mat))

        # add environment material last
        mat_treams.append(treams.Material(eps_env))

        for l in range(1, l_max + 1):
            miecoef = treams.coeffs.mie(
                l, to_np(k0[i_wl]) * np.array(radii), *zip(*mat_treams)
            )
            a_n[i_wl, l - 1] = -miecoef[0, 0] - miecoef[0, 1]
            b_n[i_wl, l - 1] = -miecoef[0, 0] + miecoef[0, 1]

    a_n = torch.as_tensor(a_n, device=device, dtype=DTYPE_COMPLEX)
    b_n = torch.as_tensor(b_n, device=device, dtype=DTYPE_COMPLEX)
    n_env = torch.as_tensor(n_env, device=device, dtype=DTYPE_COMPLEX)

    if as_dict:
        return dict(
            a_n=a_n,
            b_n=b_n,
            environment=env,
            n_env=n_env,
            device=device,
            r_enclosing=r_enclosing,
            wavelengths=wavelengths,
        )
    else:
        return a_n, b_n


def mie_crosssections_cylinder_2d(
    wavelengths: torch.Tensor,
    radii: list,
    materials: list,
    environment=None,
    m_max=10,
    device: torch.device = None,
):
    # see Bohren Huffmann, chapter 8.4
    # no autodiff!
    mie_results = mie_ab_cylinder_2d(
        wavelengths=wavelengths,
        radii=radii,
        materials=materials,
        environment=environment,
        m_max=m_max,
        device=device,
        as_dict=True,
    )
    a = mie_results["a_n"]
    b = mie_results["b_n"]
    n_env = mie_results["n_env"]
    r_enclosing = mie_results["r_enclosing"]
    wavelengths = mie_results["wavelengths"]

    k = n_env * 2 * torch.pi / torch.as_tensor(wavelengths)
    size_param = (k * r_enclosing).real

    # scattering
    Qs_mie_par = torch.abs(a[:, 0]) ** 2
    Qs_mie_perp = torch.abs(b[:, 0]) ** 2
    Qs_mie_par += 2 * torch.sum(torch.abs(a[:, 1:]) ** 2, dim=1)
    Qs_mie_perp += 2 * torch.sum(torch.abs(b[:, 1:]) ** 2, dim=1)

    Qs_mie_par *= 2 / size_param
    Qs_mie_perp *= 2 / size_param
    Qs_mie_avg = (Qs_mie_par + Qs_mie_perp) / 2

    Cs_mie_par = Qs_mie_par * 2 * r_enclosing
    Cs_mie_perp = Qs_mie_perp * 2 * r_enclosing
    Cs_mie_avg = Qs_mie_avg * 2 * r_enclosing

    # extinction
    Qe_mie_par = (a[:, 0]).real
    Qe_mie_perp = (b[:, 0]).real
    Qe_mie_par += 2 * torch.sum((a[:, 1:]), dim=1).real
    Qe_mie_perp += 2 * torch.sum((b[:, 1:]), dim=1).real

    Qe_mie_par *= 2 / size_param
    Qe_mie_perp *= 2 / size_param
    Qe_mie_avg = (Qe_mie_par + Qe_mie_perp) / 2

    Ce_mie_par = Qe_mie_par * 2 * r_enclosing
    Ce_mie_perp = Qe_mie_perp * 2 * r_enclosing
    Ce_mie_avg = Qe_mie_avg * 2 * r_enclosing

    return dict(
        Qsca_perp=Qs_mie_perp,
        Qsca_par=Qs_mie_par,
        Qsca_avg=Qs_mie_avg,
        Csca_perp=Cs_mie_perp,
        Csca_par=Cs_mie_par,
        Csca_avg=Cs_mie_avg,
        Qext_perp=Qe_mie_perp,
        Qext_par=Qe_mie_par,
        Qext_avg=Qe_mie_avg,
        Cext_perp=Ce_mie_perp,
        Cext_par=Ce_mie_par,
        Cext_avg=Ce_mie_avg,
        Qabs_perp=Qe_mie_perp - Qs_mie_perp,
        Qabs_par=Qe_mie_par - Qs_mie_par,
        Qabs_avg=Qe_mie_avg - Qs_mie_avg,
        Cabs_perp=Ce_mie_perp - Cs_mie_perp,
        Cabs_par=Ce_mie_par - Cs_mie_par,
        Cabs_avg=Ce_mie_avg - Cs_mie_avg,
        Cgeo=torch.ones_like(Cs_mie_avg) * 2 * r_enclosing,
    )


def mie_crosssections_sphere_3d(
    wavelengths: torch.Tensor,
    radii: list,
    materials: list,
    environment=None,
    l_max=10,
    device: torch.device = None,
):
    # see Bohren Huffmann, chapter 4.4
    # no autodiff!
    mie_results = mie_ab_sphere_3d(
        wavelengths=wavelengths,
        radii=radii,
        materials=materials,
        environment=environment,
        l_max=l_max,
        device=device,
        as_dict=True,
    )
    a = mie_results["a_n"]
    b = mie_results["b_n"]
    n_env = mie_results["n_env"]
    r_enclosing = mie_results["r_enclosing"]
    device = mie_results["device"]
    wavelengths = mie_results["wavelengths"]

    k = n_env * 2 * torch.pi / torch.as_tensor(wavelengths)
    size_param = (k * r_enclosing).real

    # scattering
    Qs_mie = torch.zeros(len(wavelengths), dtype=DTYPE_FLOAT, device=device)
    for l in range(l_max):
        Qs_mie += (2 * (l + 1) + 1) * (
            torch.abs(a[:, l]) ** 2 + torch.abs(b[:, l]) ** 2
        )

    Qs_mie *= 2 / size_param**2

    Cs_mie = Qs_mie * torch.pi * r_enclosing**2

    # extinction
    Qe_mie = torch.zeros(len(wavelengths), dtype=DTYPE_FLOAT, device=device)
    for l in range(l_max):
        Qe_mie += (2 * (l + 1) + 1) * (a[:, l] + b[:, l]).real

    Qe_mie *= 2 / size_param**2

    Ce_mie = Qe_mie * torch.pi * r_enclosing**2

    return dict(
        Qsca=Qs_mie,
        Csca=Cs_mie,
        Qext=Qe_mie,
        Cext=Ce_mie,
        Cgeo=torch.ones_like(Ce_mie) * torch.pi * r_enclosing**2,
    )



# -*- coding: utf-8 -*-
"""base classes for structures
"""
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX, STRUCTURE_IDS
from torchgdm.tools.misc import get_default_device
from torchgdm.tools.geometry import test_structure_distances


# --- base class structure container
class StructBase:
    """base class for structure container

    Defines the polarizabilities and self-terms
    """

    __name__ = "structure base class"

    def __init__(self, device: torch.device = None):
        """Initialization"""
        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        self.n_dim = -1  # problem dimension needs to be set by child class

        self.positions = torch.as_tensor(
            [], dtype=DTYPE_FLOAT, device=self.device
        )  # shape: (N, 3)
        self.step = torch.as_tensor(
            [], dtype=DTYPE_FLOAT, device=self.device
        )  # shape: (N, 3)

        self.mesh_normalization_factor = 1.0

        # a structure may have a pure electric or pure magnetic response
        # this can be specified, which will reduce the memory and
        # computation requirements for these scatterers
        self.evaluation_terms = ["E", "H"]

        # unique identifier
        self.id = next(STRUCTURE_IDS)

    def __add__(self, other):
        if issubclass(type(other), StructBase):
            # add a structure: try combining both.
            return self.combine(other)
        elif len(other) == 3:
            # add a 3-tuple: Try shift
            return self.translate(other)
        else:
            raise ValueError("Unknown addition.")

    def __sub__(self, other):
        if issubclass(type(other), StructBase):
            # subtract a structure:  TODO (not clear yet what to do)
            raise NotImplementedError(
                "Removing a structure from another is not implemented yet."
            )
        elif len(other) == 3:
            # subtract a 3-tuple: Try shift
            return self.translate(
                -1 * torch.as_tensor(other, dtype=DTYPE_FLOAT, device=self.device)
            )
        else:
            raise ValueError("Unknown addition.")

    def __repr__(self, verbose: bool = False):
        """description about structure"""
        out_str = " ------ base structure class - doesn't define anything yet -------"
        return out_str

    def set_device(self, device):
        """move all tensors of the class to device"""
        self.device = device
        self.positions = self.positions.to(device=device)

    # --- plot placeholder
    def plot(self):
        raise NotImplementedError("Plot not implement yet.")

    def plot_contour(self, **kwargs):
        """if not implemented: fall back to full structure plot"""
        return self.plot(**kwargs)

    def plot3d(self):
        raise NotImplementedError("3D Plot not implement yet.")

    # --- common interface for positions / step
    def get_all_positions(self) -> torch.Tensor:
        return self.positions

    def get_source_validity_radius(self) -> torch.Tensor:
        """get the radius of the validity zone of each effective source"""

        # effective pola structure: step=diameter. return step / 2
        if hasattr(self, "full_geometries"):
            _r = self.step / 2
        # discretized structure: step=cell side length. return sqrt(2) * step / 2
        else:
            _r = 2**0.5 * self.step / 2
        return _r

    # --- self-terms
    def get_selfterm_6x6(self, wavelength: float, environment) -> torch.Tensor:
        """return list of magneto-electric self-term tensors (6x6) at each meshpoint

        Args:
            wavelength (float): in nm
            environment (environment class): 3D environement class

        Returns:
            torch.Tensor: 6x6 selfterm tensor
        """
        selfterm_6x6 = torch.cat(
            [
                torch.cat(
                    [
                        self.get_selfterm_pE(wavelength, environment),
                        self.get_selfterm_pH(wavelength, environment),
                    ],
                    dim=-1,
                ),
                torch.cat(
                    [
                        self.get_selfterm_mE(wavelength, environment),
                        self.get_selfterm_mH(wavelength, environment),
                    ],
                    dim=-1,
                ),
            ],
            dim=-2,
        )
        return selfterm_6x6

    def get_selfterm_pE(self, wavelength: float, environment):
        """return list of 'EE' self-term tensors (3x3) at each meshpoint"""
        return torch.zeros(
            (len(self.positions), 3, 3), device=self.device, dtype=DTYPE_COMPLEX
        )

    def get_selfterm_mE(self, wavelength: float, environment):
        """return list of 'HE' self-term tensors (3x3) at each meshpoint"""
        return torch.zeros(
            (len(self.positions), 3, 3), device=self.device, dtype=DTYPE_COMPLEX
        )

    def get_selfterm_pH(self, wavelength: float, environment):
        """return list of 'EH' self-term tensors (3x3) at each meshpoint"""
        return torch.zeros(
            (len(self.positions), 3, 3), device=self.device, dtype=DTYPE_COMPLEX
        )

    def get_selfterm_mH(self, wavelength: float, environment):
        """return list of 'HH' self-term tensors (3x3) at each meshpoint"""
        return torch.zeros(
            (len(self.positions), 3, 3), device=self.device, dtype=DTYPE_COMPLEX
        )

    # --- polarizabilities
    def get_polarizability_6x6(self, wavelength: float, environment) -> torch.Tensor:
        """return list of magneto-electric polarizability tensors (6x6) at each meshpoint

        Args:
            wavelength (float): in nm
            environment (environment class): 3D environement class

        Returns:
            torch.Tensor: 6x6 polarizability tensor
        """
        alpha_6x6 = torch.cat(
            [
                torch.cat(
                    [
                        self.get_polarizability_pE(wavelength, environment),
                        self.get_polarizability_pH(wavelength, environment),
                    ],
                    dim=-1,
                ),
                torch.cat(
                    [
                        self.get_polarizability_mE(wavelength, environment),
                        self.get_polarizability_mH(wavelength, environment),
                    ],
                    dim=-1,
                ),
            ],
            dim=-2,
        )
        return alpha_6x6

    def get_polarizability_pmE_6x3(
        self, wavelength: float, environment
    ) -> torch.Tensor:
        alpha_pmE_6x3 = torch.cat(
            [
                self.get_polarizability_pE(wavelength, environment),
                self.get_polarizability_mE(wavelength, environment),
            ],
            dim=-2,
        )
        return alpha_pmE_6x3

    def get_polarizability_pEH_3x6(
        self, wavelength: float, environment
    ) -> torch.Tensor:
        """return list of magneto-electric self-term tensors (6x6) at each meshpoint"""
        alpha_pEH_3x6 = torch.cat(
            [
                self.get_polarizability_pE(wavelength, environment),
                self.get_polarizability_pH(wavelength, environment),
            ],
            dim=-1,
        )
        return alpha_pEH_3x6

    def get_polarizability_pmH_6x3(
        self, wavelength: float, environment
    ) -> torch.Tensor:
        alpha_pmH_6x3 = torch.cat(
            [
                self.get_polarizability_pH(wavelength, environment),
                self.get_polarizability_mH(wavelength, environment),
            ],
            dim=-2,
        )
        return alpha_pmH_6x3

    def get_polarizability_mEH_3x6(
        self, wavelength: float, environment
    ) -> torch.Tensor:
        """return list of magneto-electric self-term tensors (6x6) at each meshpoint"""
        alpha_pmH_3x6 = torch.cat(
            [
                self.get_polarizability_mE(wavelength, environment),
                self.get_polarizability_mH(wavelength, environment),
            ],
            dim=-1,
        )
        return alpha_pmH_3x6

    def get_polarizability_pE(self, wavelength: float, environment):
        """return list of EE polarizability tensors (3x3) at each meshpoint"""
        return torch.zeros(
            (len(self.positions), 3, 3), device=self.device, dtype=DTYPE_COMPLEX
        )

    def get_polarizability_mE(self, wavelength: float, environment):
        """return list of HE polarizability tensors (3x3) at each meshpoint"""
        return torch.zeros(
            (len(self.positions), 3, 3), device=self.device, dtype=DTYPE_COMPLEX
        )

    def get_polarizability_pH(self, wavelength: float, environment):
        """return list of EH polarizability tensors (3x3) at each meshpoint"""
        return torch.zeros(
            (len(self.positions), 3, 3), device=self.device, dtype=DTYPE_COMPLEX
        )

    def get_polarizability_mH(self, wavelength: float, environment):
        """return list of HH polarizability tensors (3x3) at each meshpoint"""
        return torch.zeros(
            (len(self.positions), 3, 3), device=self.device, dtype=DTYPE_COMPLEX
        )

    # - radiative correction for cross section calc. - 3D case
    def get_radiative_correction_prefactor_p(self, wavelength: float, environment):
        """return electric dipole radiative correction prefactor vectors (3) at each meshpoint"""
        if self.radiative_correction:
            if self.n_dim != 3:
                raise ValueError(
                    f"{self.n_dim}D simulation, but trying to evluate 3D radiative correction. "
                    + "Please deactivate radiative correction or implement adequate "
                    "`get_radiative_correction_prefactor_p` / `..._m`"
                )
            k0 = 2 * torch.pi / wavelength
            n_env = (
                environment.get_environment_permittivity_scalar(
                    wavelength, self.positions
                )
                ** 0.5
            )
            rad_corr_3d = (2 / 3) * k0**3 * n_env
            return torch.ones(
                (len(self.positions), 3), device=self.device, dtype=DTYPE_COMPLEX
            ) * rad_corr_3d.unsqueeze(-1)
        else:
            return torch.zeros(
                (len(self.positions), 3), device=self.device, dtype=DTYPE_COMPLEX
            )

    def get_radiative_correction_prefactor_m(self, wavelength: float, environment):
        """return magnetic dipole radiative correction prefactor vectors (3) at each meshpoint"""
        if self.radiative_correction:
            k0 = 2 * torch.pi / wavelength
            n_env = (
                environment.get_environment_permittivity_scalar(
                    wavelength, self.positions
                )
                ** 0.5
            )
            rad_corr_3d = (2 / 3) * k0**3 * n_env**3
            return torch.ones(
                (len(self.positions), 3), device=self.device, dtype=DTYPE_COMPLEX
            ) * rad_corr_3d.unsqueeze(-1)
        else:
            return torch.zeros(
                (len(self.positions), 3), device=self.device, dtype=DTYPE_COMPLEX
            )

    # --- geometry operations
    def copy(self, positions=None, rotation_angles=None):
        """(batch) copy structre to new position(s)

        optionally, the copied structures can also be batch-rotated along the `z` axis.

        Args:
            positions (list, optional): list of new positions to create copies at. If None, create a single, identical copy. Defaults to None.
            rotation_angles (list, optional): list of rotation angles for the copies. If None, keep orientations. Defaults to None.

        Returns:
            :class:`StructBase`: new structure
        """
        if type(positions) == dict:
            positions = positions["r_probe"]

        if positions is None:
            import copy

            return copy.deepcopy(self)
        else:
            # generate multiple copies, moved to `positions`
            positions = torch.as_tensor(
                positions, device=self.device, dtype=DTYPE_FLOAT
            )
            # single position: expand dim
            if len(positions.shape) == 1:
                positions = positions.unsqueeze(0)

            assert len(positions.shape) == 2
            assert positions.shape[1] == 3

            if rotation_angles is None:
                rotation_angles = torch.zeros(
                    len(positions), device=self.device, dtype=DTYPE_FLOAT
                )
            else:
                assert len(rotation_angles) == len(positions)
                assert len(rotation_angles.shape) == 1

            new_struct_list = []
            for _r, _a in zip(positions, rotation_angles):
                _struct = self.copy()
                _struct.set_center_of_mass(_r)
                if _a != 0:
                    _struct = _struct.rotate(_a)
                new_struct_list.append(_struct)

            new_struct = new_struct_list.pop(0)
            for _s in new_struct_list:
                new_struct = new_struct.combine(_s, inplace=True)

            return new_struct

    def get_geometric_crosssection(self, projection="xy"):
        """get geometric cross section the structure in nm^2

        Args:
            projection (str, optional): cartesian projection of cross section. Defaults to "xy"

        Returns:
            float: geometric cross section in nm^2
        """
        from torchgdm.tools.geometry import get_geometric_crosssection

        return get_geometric_crosssection(self, projection)

    def get_center_of_mass(self):
        """return the center of mass"""
        return torch.mean(self.positions, axis=0)

    def set_center_of_mass(self, r0_new: torch.Tensor):
        """move center of mass to new position `r0_new` (in-place)"""
        r0_new = torch.as_tensor(r0_new, device=self.positions.device)

        if len(r0_new.shape) != 1:
            if len(r0_new) not in [2, 3]:
                raise ValueError("`r0_new` needs to be (X,Y) or (X,Y,Z) tuple.")
        r0_old = self.get_center_of_mass()

        if len(r0_new) == 2:
            warnings.warn("Got 2-vector, assume xy coordinates.")
            r0_new = torch.as_tensor(
                [r0_new[0], r0_new[1], r0_old[2]], device=self.positions.device
            )

        # move
        self.positions -= r0_old  # move to origin
        self.positions += r0_new  # move to new location
        self.r0 = self.get_center_of_mass()

    def translate(self, vector):
        """return copy, moved by `vector`"""
        _shifted = self.copy()
        _shifted.positions += torch.as_tensor(
            vector, dtype=DTYPE_FLOAT, device=self.device
        ).unsqueeze(0)
        _shifted.r0 = _shifted.get_center_of_mass()
        return _shifted

    def rotate(self, alpha, center=torch.as_tensor([0.0, 0.0, 0.0]), axis="z"):
        raise NotImplementedError(
            "`rotate` is not yet implemented in the current class."
        )

    def combine(self, other, inplace=False, on_distance_violation="error"):
        """combine with a second structure (requires definition at same wavelengths!)

        Structures must be of same coupling type (electric / magnetic)

        Args:
            other (_type_): _description_
            inplace (bool, optional): Don't copy original structure, just add other structure. Can be necessary e.g. when gradients are required. Defaults to False.
            on_distance_violation (str, optional): can be "error", "warn", None (do nothing). Defaults to "error".

        Returns:
            :class:`StructBase`: new structure
        """
        if inplace:
            new_struct = self
        else:
            new_struct = self.copy()

        assert self.mesh_normalization_factor == other.mesh_normalization_factor
        assert self.radiative_correction == other.radiative_correction
        assert self.evaluation_terms == other.evaluation_terms
        assert type(self) == type(other)

        N_dist1, N_dist2 = test_structure_distances(
            self, other, on_distance_violation=on_distance_violation
        )

        new_struct.positions = torch.concatenate(
            [new_struct.positions, other.positions], dim=0
        )
        new_struct.step = torch.concatenate([new_struct.step, other.step], dim=0)
        new_struct.zeros = torch.concatenate([new_struct.zeros, other.zeros], dim=0)
        new_struct.materials = new_struct.materials + other.materials

        new_struct.r0 = self.get_center_of_mass()  # center of gravity
        return new_struct



# -*- coding: utf-8 -*-
"""
exact multipole decomposition post-processing tools for torchgdm
"""
# %%
import warnings
import copy

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm.tools.batch import get_p_m_moments, get_e0, get_h0
from torchgdm.tools.geometry import get_step_from_geometry
from torchgdm.tools.misc import get_closest_wavelength
from torchgdm.tools.special import sph_j0, sph_j1, sph_j2, sph_j3


def decomposition_exact(
    sim,
    wavelength: float,
    illumination_index: int = None,  # None: batch all illumination fields
    r0=None,
    epsilon=0.1,
    long_wavelength_approx=False,
    **kwargs,
):
    """exact multipole decomposition of the internal electric field

    Multipole decomposition of the electromagnetic field inside a nanostructure for
    electric and magnetic dipole and quadrupole moments.

    For details about the method, see:

        Alaee, R., Rockstuhl, C. & Fernandez-Corbaton, I. *An electromagnetic
        multipole expansion beyond the long-wavelength approximation.*
        Optics Communications 407, 17–21 (2018)

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): in nm
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
        r0 (torch.Tensor, optional): [x,y,z] position of multipole decomposition development. Defaults to None, in which case the center of gravity is used.
        epsilon (float, optional): positions too close to r0 will moved away from r0 by epsilon (in units of step) to avoid numerical divergence of the Bessel terms. Defaults to 0.1.
        long_wavelength_approx (bool, optional): if True, use long wavelength approximation. Defaults to False.
        kwargs: Further kwargs are ignored

    Raises:
        ValueError: Simulation has not been run yet, or magnetic polarizabilities are present, or simulation is not 3D.

    Returns:
        dict: the multipole moments. dipoles are rank-1 (3-vectors), quadrupoles are rank 2 (3x3 tensors):
            - 'ed_tot': electric dipole (full)
            - 'md': magnetic dipole
            - 'eq_tot': electric quadrupole (full)
            - 'mq': magnetic quadrupole
            - 'ed_1': electric dipole (first order)
            - 'ed_toroidal': toroidal dipole
            - 'eq1': electric quadrupole (first order)
            - 'eq_toroidal': toroidal quadrupole

    """
    # =============================================================================
    #     Exception handling
    # =============================================================================
    wavelength = get_closest_wavelength(sim, wavelength)

    if float(wavelength) not in sim.fields_inside:
        raise ValueError(
            "Error: simulation at requested wavelength not yet "
            + "evaluated. Call `sim.run()`."
        )
        
    n_dim = sim.structures[0].n_dim
    if n_dim != 3:
        raise ValueError(
            f"Dimension Error: {n_dim}D simulation, but multipoles are only "
            + "implemented for 3D simulations."
        )

    # =============================================================================
    #     preparation
    # =============================================================================
    # electric polarization density of structure
    P, M, r_p, r_m = get_p_m_moments(sim, wavelength, illumination_index)
    if len(r_m.squeeze()) > 0:
        raise ValueError(
            "Multipole decomposition does not support structures "
            + "with magnetic polarizabilties so far."
        )

    # structure. By default, use "center of mass" for multimode expansion
    if r0 is None:
        r0 = torch.mean(r_p, axis=0)
    r = r_p - r0
    r = r.unsqueeze(0)

    # avoid divergence of Bessel terms
    step = get_step_from_geometry(r_p)
    r = torch.where(
        torch.linalg.norm(r, axis=-1).unsqueeze(-1) > epsilon * step, r, epsilon
    )
    norm_r = torch.linalg.norm(r, axis=-1).unsqueeze(-1)

    # wavenumber
    eps_env_at_p = sim.environment.get_environment_permittivity_scalar(
        wavelength, r_probe=r_p
    )
    n_env_at_p = (eps_env_at_p**0.5).real.unsqueeze(0).unsqueeze(-1)
    k0 = 2.0 * torch.pi / wavelength
    k0 = torch.ones_like(n_env_at_p) * k0
    k = k0 * n_env_at_p
    kr = k * norm_r

    # precalc. bessel functions and pre-factors
    if not long_wavelength_approx:
        j0kr = sph_j0(kr)
        j1kr = sph_j1(kr) / kr
        j2kr = sph_j2(kr) / (kr**2)
        j3kr = sph_j3(kr) / (kr**3)
        f_pt = 1 / 2
        f_ptA = 3
        f_ptB = -1
        f_qe = 3
        fqe2 = 2
        fqe2A = 5
        fqe2B = -1
        fqe2C = -1
        f_m = 3 / 2
        f_qm = 15
    else:
        j0kr = torch.ones_like(kr)
        j1kr = torch.ones_like(kr)
        j2kr = torch.ones_like(kr)
        j3kr = torch.ones_like(kr)
        f_pt = 1 / 10
        f_ptA = 1
        f_ptB = -2
        f_qe = 1
        fqe2 = 1 / 14
        fqe2A = 4
        fqe2B = -5
        fqe2C = 2
        f_m = 1 / 2
        f_qm = 1

    # =============================================================================
    #     multipole calculation
    # =============================================================================
    # ----------- dipole moments
    # electric dipole
    ed_1 = torch.sum(P * j0kr, axis=1)

    ## "toroidal" dipole
    rdotp = torch.einsum(
        "kij, kij->ki", r.to(dtype=DTYPE_COMPLEX, device=sim.device), P
    ).unsqueeze(-1)
    ed_toroid = f_pt * torch.sum(
        k**2 * (f_ptA * rdotp * r + f_ptB * norm_r**2 * P) * j2kr,
        axis=1,
    )

    ed_tot = ed_1 + ed_toroid

    # magnetic dipole
    r_cross_P = torch.cross(r.to(dtype=DTYPE_COMPLEX), P, dim=-1)
    md = -1j * f_m * torch.sum(k0 * r_cross_P * j1kr, axis=1)

    # ----------- quadrupole moments
    # electric quadrupole
    eq_1 = torch.zeros((P.shape[0], 3, 3), dtype=DTYPE_COMPLEX, device=sim.device)
    eq_toroid = torch.zeros((P.shape[0], 3, 3), dtype=DTYPE_COMPLEX, device=sim.device)
    for i_a in range(3):
        for i_b in range(3):

            # diagonal term
            if i_a == i_b:
                rP_delta = rdotp
            else:
                rP_delta = torch.zeros_like(rdotp)

            # electric quadrupole
            eq_1[:, i_a, i_b] = torch.sum(
                (
                    3 * (r[..., i_a] * P[..., i_b] + r[..., i_b] * P[..., i_a])
                    - 2 * (rP_delta[..., 0])
                )
                * j1kr[..., 0],
                axis=-1,
            )

            # "toroidal" quadrupole
            eq_toroid[:, i_a, i_b] = torch.sum(
                k[..., 0] ** 2
                * (
                    fqe2A * r[..., i_a] * r[..., i_b] * rdotp[..., 0]
                    + norm_r[..., 0] ** 2
                    * (
                        fqe2B * (r[..., i_a] * P[..., i_b] + r[..., i_b] * P[..., i_a])
                        + fqe2C * rP_delta[..., 0]
                    )
                )
                * j3kr[..., 0],
                axis=-1,
            )

    eq_1 = f_qe * eq_1
    eq_toroid = f_qe * fqe2 * eq_toroid
    eq_tot = eq_1 + eq_toroid

    # magnetic quadrupole
    mq = torch.zeros((P.shape[0], 3, 3), dtype=DTYPE_COMPLEX, device=sim.device)
    for i_a in range(3):
        for i_b in range(3):
            mq[:, i_a, i_b] = torch.sum(
                k0[..., 0]
                * (
                    r[..., i_a] * r_cross_P[..., i_b]
                    + r[..., i_b] * r_cross_P[..., i_a]
                )
                * j2kr[..., 0],
                axis=-1,
            )

    mq = -1j * f_qm * mq

    return {
        "ed_1": ed_1,
        "ed_toroidal": ed_toroid,
        "ed_tot": ed_tot,
        "md": md,
        "eq_tot": eq_tot,
        "eq_1": eq_1,
        "eq_toroidal": eq_toroid,
        "mq": mq,
    }


def scs(
    sim,
    wavelength: float,
    illumination_index: int = None,  # None: batch all illumination fields
    with_toroidal=True,
    r0=None,
    epsilon=0.1,
    normalization_E0=False,
    long_wavelength_approx=False,
    **kwargs,
):
    """multipole decomposition of scattering cross section

    Returns scattering cross sections for electric and magnetic dipole and quadrupole moments.
    For details about the exact multipole formalism and scs calculation, see:

        Alaee, R., Rockstuhl, C. & Fernandez-Corbaton, I.
        *An electromagnetic multipole expansion beyond the long-wavelength approximation.*
        Optics Communications 407, 17–21 (2018)

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): in nm
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
        with_toroidal (bool, optional): whether to add toroidal moments to electric dipole and quadrupole. Defaults to True.
        r0 (torch.Tensor, optional): [x,y,z] position of multipole decomposition development. Defaults to None, in which case the center of gravity is used.
        epsilon (float, optional): positions too close to r0 will moved away from r0 by epsilon (in units of step) to avoid numerical divergence of the Bessel terms. Defaults to 0.1.
        normalization_E0 (bool, optional): Normalize to illumination amplitude at `r0`. Can be useful to get approximate results for non-plane wave illumination. Defaults to False.
        long_wavelength_approx (bool, optional): if True, use long wavelength approximation. Defaults to False.
        kwargs: Further kwargs are ignored

    Returns:
        dict:
         - 'scs_ed': electric dipole scattering cross section (in nm^2)
         - 'scs_md': magnetic dipole scattering cross section (in nm^2)
         - 'scs_eq': electric quadrupole scattering cross section (in nm^2)
         - 'scs_mq': magnetic quadrupole scattering cross section (in nm^2)
    """
    # =============================================================================
    #     Exception handling
    # =============================================================================
    from torchgdm.tools.misc import test_illumination_field_is_plane_wave

    wavelength = get_closest_wavelength(sim, wavelength)

    test_illumination_field_is_plane_wave(
        sim,
        illumination_index,
        message="Scattering cross sections are only defined for plane wave illuminations! ",
    )
    # =============================================================================
    #     scattering section calculation
    # =============================================================================
    which_ed = "ed_tot" if with_toroidal else "ed_1"
    which_md = "md"
    which_eq = "eq_tot" if with_toroidal else "eq_1"
    which_mq = "mq"

    # get dipole moments
    mpd = decomposition_exact(
        sim,
        wavelength=wavelength,
        illumination_index=illumination_index,
        r0=r0,
        epsilon=epsilon,
        long_wavelength_approx=long_wavelength_approx,
    )

    # by default, use "center of mass" for multimode expansion
    if r0 is None:
        P, M, r_p, r_m = get_p_m_moments(sim, wavelength, illumination_index)
        r0 = torch.mean(r_p, axis=0)

    # wavenumber
    eps_env_at_p = sim.environment.get_environment_permittivity_scalar(
        wavelength, r_probe=r0.unsqueeze(0)
    )[0]
    n_env_at_p = eps_env_at_p**0.5
    k0 = 2.0 * torch.pi / wavelength
    k = (k0 * n_env_at_p).to(sim.device)  # scalar

    # optional normalization to incident field intensity
    if normalization_E0:
        e_inc = get_e0(sim, wavelength, r0.unsqueeze(0), illumination_index)
        e2in = torch.sum(torch.abs(e_inc) ** 2, axis=1)  # intensity of incident field
    else:
        e2in = 1.0

    ## factor 100: cm --> m (cgs units)
    sc_factor_dp = 100 / 12 * (k0**4 / e2in).real
    sc_factor_Q = 100 / 1440 * (k0**4 / e2in).real

    scs_ed = sc_factor_dp * torch.sum(torch.abs(mpd[which_ed]) ** 2, dim=1)
    scs_md = (
        sc_factor_dp * eps_env_at_p * torch.sum(torch.abs(mpd[which_md]) ** 2, dim=1)
    )
    scs_eq = sc_factor_Q * torch.sum(torch.abs(k * mpd[which_eq]) ** 2, dim=(1, 2))
    scs_mq = (
        sc_factor_Q
        * eps_env_at_p
        * torch.sum(torch.abs(k * mpd[which_mq]) ** 2, dim=(1, 2))
    )

    return dict(scs_ed=scs_ed.real, scs_md=scs_md.real, scs_eq=scs_eq.real, scs_mq=scs_mq.real)


def ecs(
    sim,
    wavelength: float,
    illumination_index: int = None,  # None: batch all illumination fields
    with_toroidal=True,
    r0=None,
    epsilon=0.1,
    eps_dd=0.1,
    normalization_E0=False,
    long_wavelength_approx=False,
    **kwargs,
):
    """multipole decomposition of extinction cross section

    Returns extinction cross sections for electric and magnetic dipole and
    quadrupole moments of the multipole decomposition.

    *Caution:*
        The multipole extinction calculation is not entirely undiputed
        It turns out that contributions can become negative, which is
        explained by Evlyukhin et al. ([JOSA B 30, 2589 (2013)] or
        [PRB 94, 205434 (2016)]) by interactions between non-orthogonal
        multipole modes in particles that strongly break the spherical
        symmetry of this modal basis.

    For details about the extinction section of multipole moments, see:
        Evlyukhin, A. B. et al. *Multipole analysis of light scattering by
        arbitrary-shaped nanoparticles on a plane surface.*,
        JOSA B 30, 2589 (2013)

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): in nm
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
        r0 (torch.Tensor, optional): [x,y,z] position of multipole decomposition development. Defaults to None, in which case the center of gravity is used.
        epsilon (float, optional): positions too close to r0 will moved away from r0 by epsilon (in units of step) to avoid numerical divergence of the Bessel terms. Defaults to 0.1.
        eps_dd (float, optional): numerical integration step for field gradients calc. (in nm). Required for e/m quadrupoles. Defaults to 0.1.
        normalization_E0 (bool, optional): Normalize to illumination amplitude at `r0`. Can be useful to get approximate results for non-plane wave illumination. Defaults to False.
        long_wavelength_approx (bool, optional): if True, use long wavelength approximation. Defaults to False.
        kwargs: Further kwargs are ignored

    Returns:
        dict:
         - 'ecs_ed': electric dipole extinction cross section (in nm^2)
         - 'ecs_md': magnetic dipole extinction cross section (in nm^2)
         - 'ecs_eq': electric quadrupole extinction cross section (in nm^2)
         - 'ecs_mq': magnetic quadrupole extinction cross section (in nm^2)
    """

    from torchgdm.postproc.fields import field_gradient
    from torchgdm.tools.misc import test_illumination_field_is_plane_wave

    # =============================================================================
    #     Exception handling
    # =============================================================================

    wavelength = get_closest_wavelength(sim, wavelength)

    test_illumination_field_is_plane_wave(
        sim,
        illumination_index,
        message="Extinction cross sections are only defined for plane wave illuminations! ",
    )

    # =============================================================================
    #     extinction section calculation
    # =============================================================================
    which_ed = "ed_tot" if with_toroidal else "ed_1"
    which_md = "md"
    which_eq = "eq_tot" if with_toroidal else "eq_1"
    which_mq = "mq"

    # get dipole moments
    mpd = decomposition_exact(
        sim,
        wavelength=wavelength,
        illumination_index=illumination_index,
        r0=r0,
        epsilon=epsilon,
        long_wavelength_approx=long_wavelength_approx,
    )

    # by default, use "center of mass" for multimode expansion
    if r0 is None:
        P, M, r_p, r_m = get_p_m_moments(sim, wavelength, illumination_index)
        r0 = torch.mean(r_p, axis=0)

    # wavenumber
    eps_env_at_p = sim.environment.get_environment_permittivity_scalar(
        wavelength, r_probe=r0.unsqueeze(0)
    )[0]
    n_env_at_r0 = eps_env_at_p**0.5
    k0 = 2.0 * torch.pi / wavelength

    # illumination field and its gradients (index [:, 0, :] --> at single position `r0`)
    e_inc = get_e0(sim, wavelength, r0.unsqueeze(0), illumination_index)[:, 0, :]
    e_inc_cj = torch.conj(e_inc)
    h_inc = get_h0(sim, wavelength, r0.unsqueeze(0), illumination_index)[:, 0, :]
    h_inc_cj = torch.conj(h_inc)

    grad_e_inc_dict = field_gradient(
        sim, wavelength, r0.unsqueeze(0), whichfield="e_inc", delta=eps_dd
    )
    grad_e_inc = torch.concatenate(
        [grad_e_inc_dict["dfdx"], grad_e_inc_dict["dfdy"], grad_e_inc_dict["dfdz"]],
        dim=1,
    )
    grad_h_inc_dict = field_gradient(
        sim, wavelength, r0.unsqueeze(0), whichfield="h_inc", delta=eps_dd
    )
    grad_h_inc = torch.concatenate(
        [grad_h_inc_dict["dfdx"], grad_h_inc_dict["dfdy"], grad_h_inc_dict["dfdz"]],
        dim=1,
    )

    grad_e_inc_cj_T = torch.transpose(torch.conj(grad_e_inc), dim0=1, dim1=2)
    grad_h_inc_cj_T = torch.transpose(torch.conj(grad_h_inc), dim0=1, dim1=2)

    # optional normalization to incident field intensity
    if normalization_E0:
        e2in = torch.sum(torch.abs(e_inc) ** 2, axis=1)  # intensity of incident field
    else:
        e2in = 1.0

    # calculate the decomposition
    prefactor = (4.0 * torch.pi * k0 / n_env_at_r0 / e2in).real

    ecs_ed = prefactor * (torch.sum(e_inc_cj * mpd[which_ed], dim=1)).imag

    ecs_md = prefactor * (torch.sum(h_inc_cj * mpd[which_md], dim=1)).imag

    ecs_eq = (prefactor / 12.0) * (
        torch.sum(
            (torch.conj(grad_e_inc) + grad_e_inc_cj_T) * mpd[which_eq], dim=(1, 2)
        )
    ).imag

    ecs_mq = (prefactor / 6.0) * (
        torch.sum((grad_h_inc_cj_T * mpd[which_mq].transpose(-1, -2)), dim=(1, 2))
    ).imag

    # bundle results
    return dict(
        ecs_ed=ecs_ed.real, ecs_md=ecs_md.real, ecs_eq=ecs_eq.real, ecs_mq=ecs_mq.real
    )



# -*- coding: utf-8 -*-
"""
torchgdm field propgation: near-fields and far-fields
"""
# %%
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm.tools.misc import get_closest_wavelength
from torchgdm.tools.batch import batched
from torchgdm.tools.batch import add_illuminations_info_to_dict
from torchgdm.tools.batch import get_p_m_moments, get_e0, get_h0


# --- internal helper
def _check_position_devices(r1: torch.tensor, r2: torch.tensor):
    if r1.device != r2.device:
        raise ValueError(
            "Position inputs are not on same device! Is `r_probe` on same device as the simulation?"
        )


def _pos_and_src_closer_than_step(
    sim,
    r_probe: torch.Tensor,
    source_distance_steps: float = 1.0,
):
    dipole_step_radius = sim.get_source_validity_radius()
    pos_p, pos_m = sim._get_polarizable_positions_separate_p_m()
    idx_p, idx_m = sim._get_polarizable_positions_indices_separate_p_m()

    dipole_stepsizes_p = torch.index_select(dipole_step_radius, 0, idx_p)
    dipole_stepsizes_m = torch.index_select(dipole_step_radius, 0, idx_m)

    # p-sources too close to r_probe
    all_dist_vec_p = r_probe.unsqueeze(1) - pos_p.unsqueeze(0)
    all_dist_p = torch.linalg.norm(all_dist_vec_p, dim=-1)
    mask_inner_p = all_dist_p <= (source_distance_steps * dipole_stepsizes_p)

    # m-sources too close to r_probe
    all_dist_vec_m = r_probe.unsqueeze(1) - pos_m.unsqueeze(0)
    all_dist_m = torch.linalg.norm(all_dist_vec_m, dim=-1)
    mask_inner_m = all_dist_m <= (source_distance_steps * dipole_stepsizes_m)

    # return masks for probe locations inside and outside
    return (mask_inner_p, mask_inner_m)


def _eval_Gpm(func_Gp, func_Gm, r_probe, r_p, r_m, wavelength):
    _check_position_devices(r_probe, r_p)

    # Green's tensors for propagation
    Gp = func_Gp(r_probe.unsqueeze(1), r_p.unsqueeze(0), wavelength=wavelength)
    Gm = func_Gm(r_probe.unsqueeze(1), r_m.unsqueeze(0), wavelength=wavelength)
    return Gp, Gm


# ******************************************************************
# main field calculation routines
# ******************************************************************
# --- private functions internally used by high-level interface
@batched(
    batch_kwarg="r_probe",
    arg_position=2,
    out_dim_batch=1,
    default_batch_size=1024,
    title="batched nearfield calculation",
)
def _nearfield(
    sim,
    wavelength: float,
    r_probe: torch.Tensor,
    illumination_index: int = None,  # None: batch all illumination fields
    source_distance_steps: float = 1.0,
    interpolation_step_range: float = 1.0,  # internal field interpolation
    **kwargs,
):
    wavelength = get_closest_wavelength(sim, wavelength)

    # test r_probe consistency
    r_probe = r_probe.to(sim.device)
    assert len(r_probe.shape) == 2
    assert r_probe.shape[1] == 3

    # source positions and dipole moments
    p, m, pos_p, pos_m = get_p_m_moments(sim, wavelength, illumination_index)

    # Green's tensors for propagation
    G_EHp_6x3, G_EHm_6x3 = _eval_Gpm(
        sim.environment.get_G_EHp_6x3,
        sim.environment.get_G_EHm_6x3,
        r_probe,
        pos_p,
        pos_m,
        wavelength,
    )

    # scattered fields from all sources
    all_EHp = torch.matmul(
        G_EHp_6x3.unsqueeze(1), p.unsqueeze(0).unsqueeze(-1)
    ).swapaxes(1, 2)[..., 0]
    all_EHm = torch.matmul(
        G_EHm_6x3.unsqueeze(1), m.unsqueeze(0).unsqueeze(-1)
    ).swapaxes(1, 2)[..., 0]

    # identify and ignore probe positions too close to source points
    if source_distance_steps is not None:
        mask_inner_p, mask_inner_m = _pos_and_src_closer_than_step(
            sim,
            r_probe,
            source_distance_steps,
        )
        # set diverging tensors zero
        all_EHp[mask_inner_p] = 0
        all_EHm[mask_inner_m] = 0

    # superpose
    scat_EHp = torch.sum(all_EHp, axis=1).swapaxes(0, 1)
    scat_EHm = torch.sum(all_EHm, axis=1).swapaxes(0, 1)
    scat_EH = scat_EHp + scat_EHm

    # separate combined fields into E and H; creatae copies to allow aitograd
    e_sca, h_sca = torch.chunk(scat_EH, 2, dim=2)
    e_sca = e_sca.clone()
    h_sca = h_sca.clone()

    # calc. illumination and total fields
    e_inc = get_e0(sim, wavelength, r_probe, illumination_index)
    h_inc = get_h0(sim, wavelength, r_probe, illumination_index)

    e_tot = e_sca + e_inc
    h_tot = h_sca + h_inc

    # treat internal fields separately
    _e_tot_in, mask_r_e_in = _internal_fields_e(
        sim,
        wavelength=wavelength,
        r_probe=r_probe,
        illumination_index=illumination_index,
        interpolation_step_range=interpolation_step_range,
    )
    e_tot[:, mask_r_e_in] = _e_tot_in
    e_sca[:, mask_r_e_in] = e_tot[:, mask_r_e_in] - e_inc[:, mask_r_e_in]

    _h_tot_in, mask_r_h_in = _internal_fields_h(
        sim,
        wavelength=wavelength,
        r_probe=r_probe,
        illumination_index=illumination_index,
        interpolation_step_range=interpolation_step_range,
    )
    h_tot[:, mask_r_h_in] = _h_tot_in
    h_sca[:, mask_r_h_in] = h_tot[:, mask_r_h_in] - h_inc[:, mask_r_h_in]

    # wrap up in dict and add information on illumination fields
    res_dict = dict(
        e_sca=e_sca, e_tot=e_tot, e_inc=e_inc, h_sca=h_sca, h_tot=h_tot, h_inc=h_inc
    )
    
    return res_dict


def _internal_fields_e(
    sim,
    wavelength: float,
    r_probe: torch.Tensor,
    illumination_index: int = None,  # None: batch all illumination fields
    interpolation_step_range=1.0,
    min_step_radius_distance=1.0,
    epsilon=1e-2,
):
    wavelength = sim.get_closest_wavelength(wavelength)
    dipole_step_radius = sim.get_source_validity_radius()
    pos_p, pos_m = sim._get_polarizable_positions_separate_p_m()
    idx_p, idx_m = sim._get_polarizable_positions_indices_separate_p_m()
    step_radius_p = torch.index_select(dipole_step_radius, 0, idx_p)

    # all distances between probe and sources
    all_dist_vec = r_probe.unsqueeze(1) - pos_p.unsqueeze(0)
    all_dist = torch.linalg.norm(all_dist_vec, dim=-1)

    # internal positions
    mask_inner = all_dist <= step_radius_p / min_step_radius_distance + epsilon
    mask_pos_in = torch.logical_not(torch.all(torch.logical_not(mask_inner), dim=1))

    # get weighted positions of internal field cells close to singular r_probe
    mask_interpol = all_dist <= (interpolation_step_range * step_radius_p)
    all_weights = 1 / (all_dist + epsilon)
    all_weights[torch.logical_not(mask_interpol)] = 0
    all_weights = all_weights / all_weights.sum(keepdim=True, dim=-1)  # normalize
    all_weights[torch.logical_not(mask_interpol)] = 0
    all_weights = all_weights[mask_pos_in]

    # get internal fields
    e_in_dat = sim._get_polarizablefields_e(wavelength)
    if illumination_index is not None:
        e_in_dat = e_in_dat[illumination_index].unsqueeze(0)

    # weighted average of internal fields close to singular r_probe
    e_in_weighted_sum = e_in_dat.unsqueeze(1) * all_weights.unsqueeze(0).unsqueeze(-1)
    e_in_interpolated = torch.sum(e_in_weighted_sum, axis=2)

    return e_in_interpolated, mask_pos_in


def _internal_fields_h(
    sim,
    wavelength: float,
    r_probe: torch.Tensor,
    illumination_index: int = None,  # None: batch all illumination fields
    interpolation_step_range=1.0,
    min_step_radius_distance=1.0,
    epsilon=1e-2,
):
    wavelength = get_closest_wavelength(sim, wavelength)

    dipole_step_radius = sim.get_source_validity_radius()
    pos_p, pos_m = sim._get_polarizable_positions_separate_p_m()
    idx_p, idx_m = sim._get_polarizable_positions_indices_separate_p_m()
    step_radius_m = torch.index_select(dipole_step_radius, 0, idx_m)

    # all distances between probe and sources
    all_dist_vec = r_probe.unsqueeze(1) - pos_m.unsqueeze(0)
    all_dist = torch.linalg.norm(all_dist_vec, dim=-1)

    # internal positions
    mask_inner = all_dist <= step_radius_m * min_step_radius_distance + epsilon
    mask_pos_in = torch.logical_not(torch.all(torch.logical_not(mask_inner), dim=1))

    # get weighted positions of internal field cells close to singular r_probe
    mask_interpol = all_dist <= (interpolation_step_range * step_radius_m)
    all_weights = 1 / (all_dist + epsilon)
    all_weights[torch.logical_not(mask_interpol)] = 0
    all_weights = all_weights / all_weights.sum(keepdim=True, dim=-1)  # normalize
    all_weights[torch.logical_not(mask_interpol)] = 0
    all_weights = all_weights[mask_pos_in]

    # get internal fields
    h_in_dat = sim._get_polarizablefields_h(wavelength)
    if illumination_index is not None:
        h_in_dat = h_in_dat[illumination_index].unsqueeze(0)

    # weighted average of internal fields close to singular r_probe
    h_in_weighted_sum = h_in_dat.unsqueeze(1) * all_weights.unsqueeze(0).unsqueeze(-1)
    h_in_interpolated = torch.sum(h_in_weighted_sum, axis=2)

    return h_in_interpolated, mask_pos_in


# @batched(  # not necessary since `_nearfield` is already batched
#     batch_kwarg="r_probe", arg_position=2, out_dim_batch=1, default_batch_size=1024
# )
def _poynting(
    sim,
    wavelength: float,
    r_probe: torch.Tensor,
    illumination_index: int = None,  # None: batch all illumination fields
    source_distance_steps: float = None,
    whichfield="tot",
    **kwargs,
):
    nf_results = _nearfield(
        sim=sim,
        wavelength=wavelength,
        r_probe=r_probe,
        illumination_index=illumination_index,
        source_distance_steps=source_distance_steps,
        pgr_bar_title="batched Poynting calculation",
        **kwargs,
    )

    if whichfield.lower() == "tot":
        e, h = nf_results["e_tot"], nf_results["h_tot"]
    elif whichfield.lower() == "sca":
        e, h = nf_results["e_sca"], nf_results["h_sca"]
    elif whichfield.lower() in ["in", "inc", "0", "zero"]:
        e, h = nf_results["e_inc"], nf_results["h_inc"]

    S = torch.cross(torch.conj(e), h, dim=-1)

    # wrap up in dict and add information on illumination fields
    res_dict = dict(poynting=S)
    return res_dict


# @batched(  # not necessary since `_nearfield` is already batched
#     batch_kwarg="r_probe", arg_position=2, out_dim_batch=1, default_batch_size=1024
# )
def _chirality(
    sim,
    wavelength: float,
    r_probe: torch.Tensor,
    illumination_index: int = None,  # None: batch all illumination fields
    source_distance_steps: float = None,
    whichfield="tot",
    **kwargs,
):
    nf_results = _nearfield(
        sim=sim,
        wavelength=wavelength,
        r_probe=r_probe,
        illumination_index=illumination_index,
        source_distance_steps=source_distance_steps,
        pgr_bar_title="batched chirality calculation",
        **kwargs,
    )
    if whichfield.lower() == "tot":
        e, h = nf_results["e_tot"], nf_results["h_tot"]
    elif whichfield.lower() == "sca":
        e, h = nf_results["e_sca"], nf_results["h_sca"]
    elif whichfield.lower() in ["in", "inc", "0", "zero"]:
        e, h = nf_results["e_inc"], nf_results["h_inc"]

    # Meinzer et al, PRB 88, 041407, 2013: C ~ Im(conj(E) * B)
    C = -1 * torch.sum(torch.multiply(torch.conj(e), h), dim=-1).imag

    # wrap up in dict and add information on illumination fields
    res_dict = dict(chirality=C)
    return res_dict


@batched(
    batch_kwarg="r_probe",
    arg_position=2,
    out_dim_batch=1,
    default_batch_size=1024,
    title="batched farfield calculation",
)
def _farfield(
    sim,
    wavelength: float,
    r_probe: torch.Tensor,
    illumination_index: int = None,  # None: batch all illumination fields
    source_distance_steps: float = None,
    **kwargs,
):
    """calulate the farfield

    Note: in the far-field, the e/m waves are transverse. E, H and k are in a fixed, orthogonal relation.
    Therefore `farfield` calculates only the electric field.

    Args:
        sim (SimulationBase):               simulation
        wavelength (float):                 evaluation wavelength (nm)
        r_probe (torch.Tensor):             positions to evaluate (sufficiently far away)
        illumination_index (int, optional): Optional index of evaluation illumination field.
                                            If None, batch-eval all. Defaults to None.

    Returns:
        dict: scattered ("ff_e_sca"), total ("ff_e_tot") and illumination ("ff_e_inc") e-fields
    """
    wavelength = get_closest_wavelength(sim, wavelength)

    # test r_probe consistency
    r_probe = r_probe.to(sim.device)
    assert len(r_probe.shape) == 2
    assert r_probe.shape[1] == 3

    # source positions and dipole moments
    p, m, r_p, r_m = get_p_m_moments(sim, wavelength, illumination_index)

    # Green's tensors for propagation
    G_Ep_3x3, G_Em_3x3 = _eval_Gpm(
        sim.environment.get_G_Ep_farfield,
        sim.environment.get_G_Em_farfield,
        r_probe,
        r_p,
        r_m,
        wavelength,
    )

    # scattered fields from all sources
    all_Ep = torch.matmul(G_Ep_3x3.unsqueeze(1), p.unsqueeze(0).unsqueeze(-1)).swapaxes(
        1, 2
    )[..., 0]
    all_Em = torch.matmul(G_Em_3x3.unsqueeze(1), m.unsqueeze(0).unsqueeze(-1)).swapaxes(
        1, 2
    )[..., 0]

    # ignore source points too close to probe position
    if source_distance_steps is not None:
        all_Ep, all_Em = _pos_and_src_closer_than_step(
            sim, all_Ep, all_Em, r_probe, source_distance_steps
        )

    # superpose
    scat_Ep = torch.sum(all_Ep, axis=1).swapaxes(0, 1)
    scat_Em = torch.sum(all_Em, axis=1).swapaxes(0, 1)
    e_sca = scat_Ep + scat_Em

    # illumination fields, total fields
    e_inc = get_e0(sim, wavelength, r_probe, illumination_index)
    e_tot = e_sca + e_inc

    # wrap up in dict and add information on illumination fields
    res_dict = dict(ff_e_sca=e_sca, ff_e_tot=e_tot, ff_e_inc=e_inc)
    return res_dict


# --- alternative e-only / h-only:
@batched(
    batch_kwarg="r_probe",
    arg_position=2,
    out_dim_batch=1,
    default_batch_size=1024,
    title="batched nearfield calculation (E only)",
)
def _nearfield_e(
    sim,
    wavelength: float,
    r_probe: torch.Tensor,
    illumination_index: int = None,  # None: batch all illumination fields
    source_distance_steps: float = None,
    **kwargs,
):
    """calculate E-field only.
    Note: this does not correctly process positions inside a structure

    further kwargs are ignored
    """

    wavelength = get_closest_wavelength(sim, wavelength)

    # test r_probe consistency
    r_probe = r_probe.to(sim.device)
    assert len(r_probe.shape) == 2
    assert r_probe.shape[1] == 3

    # source positions and dipole moments
    p, m, r_p, r_m = get_p_m_moments(sim, wavelength, illumination_index)

    # Green's tensors for propagation
    G_Ep_3x3, G_Em_3x3 = _eval_Gpm(
        sim.environment.get_G_Ep,
        sim.environment.get_G_Em,
        r_probe,
        r_p,
        r_m,
        wavelength,
    )

    # scattered fields from all sources
    all_Ep = torch.matmul(G_Ep_3x3.unsqueeze(1), p.unsqueeze(0).unsqueeze(-1)).swapaxes(
        1, 2
    )[..., 0]
    all_Em = torch.matmul(G_Em_3x3.unsqueeze(1), m.unsqueeze(0).unsqueeze(-1)).swapaxes(
        1, 2
    )[..., 0]

    # identify and ignore probe positions too close to source points
    if source_distance_steps is not None:
        mask_inner_p, mask_inner_m = _pos_and_src_closer_than_step(
            sim, r_probe, source_distance_steps
        )
        # set diverging tensors zero
        all_Ep[mask_inner_p] = 0
        all_Em[mask_inner_m] = 0

    # superpose
    scat_Ep = torch.sum(all_Ep, axis=1).swapaxes(0, 1)
    scat_Em = torch.sum(all_Em, axis=1).swapaxes(0, 1)
    e_sca = scat_Ep + scat_Em

    # illumination fields, total fields
    e_inc = get_e0(sim, wavelength, r_probe, illumination_index)
    e_tot = e_sca + e_inc

    # wrap up in dict and add information on illumination fields
    res_dict = dict(e_sca=e_sca, e_tot=e_tot, e_inc=e_inc)
    return res_dict


@batched(
    batch_kwarg="r_probe",
    arg_position=2,
    out_dim_batch=1,
    default_batch_size=1024,
    title="batched nearfield calculation (H only)",
)
def _nearfield_h(
    sim,
    wavelength: float,
    r_probe: torch.Tensor,
    illumination_index: int = None,  # None: batch all illumination fields
    source_distance_steps: float = None,
    **kwargs,
):
    """calculate H-field only.
    Note: this does not correctly process positions inside a structure

    further kwargs are ignored
    """
    wavelength = get_closest_wavelength(sim, wavelength)

    # test r_probe consistency
    r_probe = r_probe.to(sim.device)
    assert len(r_probe.shape) == 2
    assert r_probe.shape[1] == 3

    # source positions and dipole moments
    p, m, r_p, r_m = get_p_m_moments(sim, wavelength, illumination_index)

    # Green's tensors for propagation
    G_Hp_3x3, G_Hm_3x3 = _eval_Gpm(
        sim.environment.get_G_Hp,
        sim.environment.get_G_Hm,
        r_probe,
        r_p,
        r_m,
        wavelength,
    )

    # scattered fields from all sources
    all_Hp = torch.matmul(G_Hp_3x3.unsqueeze(1), p.unsqueeze(0).unsqueeze(-1)).swapaxes(
        1, 2
    )[..., 0]
    all_Hm = torch.matmul(G_Hm_3x3.unsqueeze(1), m.unsqueeze(0).unsqueeze(-1)).swapaxes(
        1, 2
    )[..., 0]

    # identify and ignore probe positions too close to source points
    if source_distance_steps is not None:
        mask_inner_p, mask_inner_m = _pos_and_src_closer_than_step(
            sim, r_probe, source_distance_steps
        )
        # set diverging tensors zero
        all_Hp[mask_inner_p] = 0
        all_Hm[mask_inner_m] = 0

    # superpose
    scat_Hp = torch.sum(all_Hp, axis=1).swapaxes(0, 1)
    scat_Hm = torch.sum(all_Hm, axis=1).swapaxes(0, 1)
    h_sca = scat_Hp + scat_Hm

    # illumination fields, total fields
    h_inc = get_h0(sim, wavelength, r_probe, illumination_index)
    h_tot = h_sca + h_inc

    # wrap up in dict and add information on illumination fields
    res_dict = dict(h_sca=h_sca, h_tot=h_tot, h_inc=h_inc)
    return res_dict


# --- public functions of high-level API
@batched(batch_kwarg="r_probe", arg_position=2, out_dim_batch=1, default_batch_size=256,title="batched field gradient calculation")
def _field_grad(
    sim,
    wavelength: float,
    r_probe: torch.Tensor,
    illumination_index: int = None,  # None: batch all illumination fields
    source_distance_steps: float = None,
    delta=0.1,
    whichfield="e_tot",
    **kwargs,
):
    """nearfield gradient distribution inside or in proximity of a nanostructure"""
    # =============================================================================
    #     Exception handling
    # =============================================================================
    wavelength = get_closest_wavelength(sim, wavelength)

    # if r_probe is None:
    #     r_probe = sim.get_all_positions()

    if type(r_probe) == dict:
        r_probe = r_probe["r_probe"]

    # move probe positions to same device as simulation
    r_probe = torch.as_tensor(r_probe, device=sim.device, dtype=DTYPE_FLOAT)
    if len(r_probe.shape) == 1:
        r_probe = r_probe.unsqueeze(0)
    assert len(r_probe.shape) == 2
    assert r_probe.shape[1] == 3

    # =============================================================================
    #     preparation
    # =============================================================================
    whichfield = whichfield.lower()
    if whichfield not in ["e_tot", "e_sca", "e_inc", "h_tot", "h_sca", "h_inc"]:
        raise ValueError(
            "Wrong argument for `field`. Must be one of "
            + "['e_tot', 'e_sca', 'e_inc', 'h_tot', 'h_sca', 'h_inc']."
        )

    kw = dict(
        sim=sim,
        wavelength=wavelength,
        illumination_index=illumination_index,
    )

    if whichfield.lower() == "e_inc":

        def get_e0_wrap(**kwargs):
            return dict(e_inc=get_e0(**kwargs))

        field_func = get_e0_wrap

    elif whichfield.lower() == "h_inc":

        def get_h0_wrap(**kwargs):
            return dict(h_inc=get_h0(**kwargs))

        field_func = get_h0_wrap

    else:
        field_func = _nearfield
        kw["source_distance_steps"] = source_distance_steps

    # =============================================================================
    #     calc field gradients
    # =============================================================================
    e_inc = get_e0(sim, wavelength, r_probe, illumination_index, **kwargs)
    h_inc = get_h0(sim, wavelength, r_probe, illumination_index, **kwargs)

    ## --- d/dx
    r_probe_px = torch.clone(r_probe)  # Xmap+delta
    r_probe_mx = torch.clone(r_probe)  # Xmap-delta
    r_probe_px.T[0] += delta
    r_probe_mx.T[0] -= delta
    F_px = field_func(r_probe=r_probe_px, **kw)[whichfield]
    F_mx = field_func(r_probe=r_probe_mx, **kw)[whichfield]

    dFdx = (F_px - F_mx) / (2 * delta)

    ## --- d/dy
    r_probe_py = torch.clone(r_probe)  # Ymap+delta
    r_probe_my = torch.clone(r_probe)  # Ymap-delta
    r_probe_py.T[1] += delta
    r_probe_my.T[1] -= delta
    F_py = field_func(r_probe=r_probe_py, **kw)[whichfield]
    F_my = field_func(r_probe=r_probe_my, **kw)[whichfield]

    dFdy = (F_py - F_my) / (2 * delta)

    ## --- d/dz
    r_probe_pz = torch.clone(r_probe)  # Zmap+delta
    r_probe_mz = torch.clone(r_probe)  # Zmap-delta
    r_probe_pz.T[2] += delta
    r_probe_mz.T[2] -= delta
    F_pz = field_func(r_probe=r_probe_pz, **kw)[whichfield]
    F_mz = field_func(r_probe=r_probe_mz, **kw)[whichfield]

    dFdz = (F_pz - F_mz) / (2 * delta)

    # wrap up in dict and add information on illumination fields
    res_dict = dict(dfdx=dFdx, dfdy=dFdy, dfdz=dFdz)
    return res_dict


# --- high-level interface
def nf(
    sim,
    wavelength: float,
    r_probe: dict = None,
    illumination_index: int = None,  # None: batch all illumination fields
    progress_bar=True,
    **kwargs,
):
    """nearfield (electric and magnetic) at positions `r_probe`

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): evaluation wavelength (in nm)
        r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`).  Defaults to None, in which case all positions of the simulation structures are used
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
        progress_bar (bool, optional): Show progress bar. Defaults to True.

    Returns:
        dict: contains results for total, scattered and incident fields in instances of :class:`torchgdm.Field`
    """
    from torchgdm.field import Field

    if r_probe is None:
        r_probe = sim.get_all_positions()
        _r = r_probe
        warnings.warn(
            "No positions given. Falling back to internal fields. "
            + "For internal field intensities, you may use the "
            + "faster `integrated_nf_intensity_e_inside`."
        )
    elif type(r_probe) == dict:
        _r = r_probe["r_probe"]
    else:
        r_probe = torch.as_tensor(r_probe, device=sim.device, dtype=DTYPE_FLOAT)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)
        _r = r_probe

    nf_res = _nearfield(
        sim=sim,
        wavelength=wavelength,
        r_probe=_r,
        illumination_index=illumination_index,
        progress_bar=progress_bar,
        **kwargs,
    )
    res_dict = {}

    res_dict["sca"] = Field(r_probe, efield=nf_res["e_sca"], hfield=nf_res["h_sca"])
    res_dict["tot"] = Field(r_probe, efield=nf_res["e_tot"], hfield=nf_res["h_tot"])
    res_dict["inc"] = Field(r_probe, efield=nf_res["e_inc"], hfield=nf_res["h_inc"])
    res_dict["wavelength"] = wavelength

    # add information on illumination field config
    res_dict = add_illuminations_info_to_dict(
        sim, res_dict, illumination_index=illumination_index
    )

    return res_dict


def field_gradient(
    sim,
    wavelength: float,
    r_probe: torch.Tensor,
    illumination_index: int = None,  # None: batch all illumination fields
    source_distance_steps: float = None,
    delta=0.1,
    whichfield="e_tot",
    **kwargs,
):
    """nearfield gradient distribution inside or in proximity of a nanostructure

    Calculate field-gradients (positions defined by `r_probe`).
    pytorch AD is not efficient for gradients of functions R^n --> R^m, with $n \\sim m >> 1$.
    Therefore, here numerical derivatives are calculated via center differences.

    Based on the original implementation in pyGDM by C. Majorel.

    Warning: The current implementation is not memory efficient, since all fields are calculated, even though not required.

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): in nm
        r_probe (torch.Tensor): tuple (x,y,z) or list of 3-lists/-tuples to evaluate field gradients.
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
        delta (float, optional):  differential step for numerical center-derivative (in nanometers). Defaults to 0.1.
        whichfield (str, optional): fields to calculate the gradient for. One of ["e_sca","e_tot","e_inc", "h_sca","h_tot,"h_inc], . Defaults to "e_tot".

    Raises:
        ValueError: _description_

    Returns:
        3 lists of 3-tuples [dAx, dAy, dAz] (complex): dAj are the differential terms:
            - idx [0] = dE/dx = [dEx/dx, dEy/dx, dEz/dx]
            - idx [1] = dE/dy = [dEx/dy, dEy/dy, dEz/dy]
            - idx [2] = dE/dz = [dEx/dz, dEy/dz, dEz/dz]
    """
    if type(r_probe) == dict:
        _r_probe = r_probe["r_probe"]
    else:
        _r_probe = r_probe

    grad_f_dict = _field_grad(
        sim=sim,
        wavelength=wavelength,
        r_probe=_r_probe,
        illumination_index=illumination_index,
        source_distance_steps=source_distance_steps,
        delta=delta,
        whichfield=whichfield,
        **kwargs,
    )

    return grad_f_dict


def ff(
    sim,
    wavelength: float,
    r_probe: dict = None,
    illumination_index: int = None,  # None: batch all illumination fields
    progress_bar=True,
    **kwargs,
):
    """far-field (electric and magnetic) at positions `r_probe`

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): evaluation wavelength (in nm)
        r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`). Defaults to None, in which case full hemisphere coordinates are used
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
        progress_bar (bool, optional): Show progress bar. Defaults to True.

    Returns:
        dict: contains results for total, scattered and incident fields in instances of :class:`torchgdm.Field`
    """

    from torchgdm.field import Field

    if r_probe is None:
        from torchgdm.tools.geometry import coordinate_map_2d_spherical

        r_probe = coordinate_map_2d_spherical()  # full hemisphere
        _r = r_probe["r_probe"]
    elif type(r_probe) == dict:
        _r = r_probe["r_probe"]
    else:
        r_probe = torch.as_tensor(r_probe, device=sim.device, dtype=DTYPE_FLOAT)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)
        _r = r_probe

    ff_res = _farfield(
        sim=sim,
        wavelength=wavelength,
        r_probe=_r,
        illumination_index=illumination_index,
        progress_bar=progress_bar,
        **kwargs,
    )
    res_dict = {}

    res_dict["sca"] = Field(r_probe, efield=ff_res["ff_e_sca"])
    res_dict["tot"] = Field(r_probe, efield=ff_res["ff_e_tot"])
    res_dict["inc"] = Field(r_probe, efield=ff_res["ff_e_inc"])
    res_dict["wavelength"] = wavelength

    # add information on illumination field config
    res_dict = add_illuminations_info_to_dict(
        sim, res_dict, illumination_index=illumination_index
    )

    return res_dict


def integrated_nf_intensity_e(
    sim,
    wavelength: float,
    r_probe: dict = None,
    illumination_index: int = None,  # None: batch all illumination fields
    **kwargs,
):
    """integrated electric near-field intensity at positions `r_probe`

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): evaluation wavelength (in nm)
        r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`). Defaults to None, in which case all positions of the simulation structures are used
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

    Returns:
        dict: contains results for total, scattered and incident integrated E-field intensity
    """
    r_nf = nf(sim, wavelength, r_probe, illumination_index, **kwargs)

    res_dict = {}

    res_dict["wavelength"] = wavelength
    res_dict["sca"] = r_nf["sca"].get_integrated_efield_intensity()
    res_dict["tot"] = r_nf["tot"].get_integrated_efield_intensity()
    res_dict["inc"] = r_nf["inc"].get_integrated_efield_intensity()

    # add information on illumination field config
    res_dict = add_illuminations_info_to_dict(
        sim, res_dict, illumination_index=illumination_index
    )

    return res_dict


def integrated_nf_intensity_e_inside(
    sim,
    wavelength: float,
    illumination_index: int = None,  # None: batch all illumination fields
    **kwargs,
):
    """rapid function to calculate average internal total electric field intensity"""
    from torchgdm import to_np

    wl = float(to_np(wavelength))
    I_e = sim.fields_inside[wl].get_integrated_efield_intensity()
    I_e = I_e / len(sim.get_all_positions())

    res_dict = {}
    res_dict["wavelength"] = wavelength
    res_dict["tot"] = I_e

    # add information on illumination field config
    res_dict = add_illuminations_info_to_dict(
        sim, res_dict, illumination_index=illumination_index
    )

    return res_dict


def integrated_nf_intensity_h(
    sim,
    wavelength: float,
    r_probe: dict = None,
    illumination_index: int = None,  # None: batch all illumination fields
    **kwargs,
):
    """integrated magnetic near-field intensity at positions `r_probe`

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): evaluation wavelength (in nm)
        r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`). Defaults to None, in which case all positions of the simulation structures are used
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

    Returns:
        dict: contains results for total, scattered and incident integrated H-field intensity
    """
    r_nf = nf(sim, wavelength, r_probe, illumination_index, **kwargs)

    res_dict = {}

    res_dict["wavelength"] = wavelength
    res_dict["sca"] = r_nf["sca"].get_integrated_hfield_intensity()
    res_dict["tot"] = r_nf["tot"].get_integrated_hfield_intensity()
    res_dict["inc"] = r_nf["inc"].get_integrated_hfield_intensity()

    # add information on illumination field config
    res_dict = add_illuminations_info_to_dict(
        sim, res_dict, illumination_index=illumination_index
    )

    return res_dict


def integrated_nf_intensity_h_inside(
    sim,
    wavelength: float,
    illumination_index: int = None,  # None: batch all illumination fields
    **kwargs,
):
    """rapid function to calculate average internal total magnetic field intensity"""
    from torchgdm import to_np

    wl = float(to_np(wavelength))
    I_h = sim.fields_inside[wl].get_integrated_hfield_intensity()
    I_h = I_h / len(sim.get_all_positions())

    res_dict = {}
    res_dict["wavelength"] = wavelength
    res_dict["tot"] = I_h

    # add information on illumination field config
    res_dict = add_illuminations_info_to_dict(
        sim, res_dict, illumination_index=illumination_index
    )

    return res_dict


def integrated_ff_intensity(
    sim,
    wavelength: float,
    r_probe: dict = None,
    illumination_index: int = None,  # None: batch all illumination fields
    **kwargs,
):
    """integrated far-field intensity (E-field) at positions `r_probe`

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): evaluation wavelength (in nm)
        r_probe (dict, optional): probe locations describing dict (gen. from `tools.geometry`). Defaults to None, in which case full hemisphere coordinates are used
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.

    Returns:
        dict: contains results for total, scattered and incident integrated far-field E-field intensity
    """
    r_ff = ff(sim, wavelength, r_probe, illumination_index, **kwargs)

    res_dict = {}

    res_dict["wavelength"] = wavelength
    res_dict["sca"] = r_ff["sca"].get_integrated_efield_intensity()
    res_dict["tot"] = r_ff["tot"].get_integrated_efield_intensity()
    res_dict["inc"] = r_ff["inc"].get_integrated_efield_intensity()

    # add information on illumination field config
    res_dict = add_illuminations_info_to_dict(
        sim, res_dict, illumination_index=illumination_index
    )

    return res_dict


# --- testing
if __name__ == "__main__":
    import time
    import pickle
    import copy

    import numpy as np
    import matplotlib.pyplot as plt
    import torch

    # self-imports
    import torchgdm as tg

    # def interaction_matrix_EE(geo, wavelength, selfterms, alpha, eps_env):
    device = "cpu"
    pygdmmethod = "scipyinv"

    device = "cuda:0"
    pygdmmethod = "cupy"

    # - environment
    eps_env = 1.0
    mat_env = tg.materials.MatConstant(eps_env, device=device)
    env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env, device=device)

    # - illumination field(s)
    wavelengths = torch.tensor([500.0])
    # wavelengths = torch.linspace(500.0, 900., 100)

    e_inc_list = [
        tg.env.freespace_3d.inc_fields.PlaneWave(
            e0p=1.0, e0s=0.0, inc_angle=inc, device=device
        )  # X-polarization, incidence from top
        for inc in torch.linspace(0, np.pi, 10)
    ]

    # - structure
    l = 8
    w = 8
    h = 8

    # l = 3;w = 1;h = 1

    step = 25.0
    eps_mat = 12.0
    mat_struct = tg.materials.MatConstant(eps=eps_mat, device=device)
    struct = tg.struct.volume.StructDiscretizedCubic3D(
        tg.struct.volume.cuboid(l, w, h), step=step, materials=mat_struct, device=device
    )

    sim = tg.Simulation(
        structures=[struct],
        environment=env,
        illumination_fields=e_inc_list,
        wavelengths=wavelengths,
        device=device,
        # linearsystem=linearsystem.LinearSystemFullInverse(device=device)
    )
    sim.run(calc_missing=True)

    # %% - scattered nearfield
    if 1:
        from pyGDM2 import visu as pygdm_visu

        wavelength = float(wavelengths[0])
        surface = tg.tools.geometry.coordinate_map_2d_square(200, 21, 300)
        r_probe = surface["r_probe"]

        t0 = time.time()
        nf_results = nf(sim, wavelength, surface)

        F = nf_results["tot"]
        e_tot = F.efield

        t1 = time.time()
        print("ff {:}pts, e and h: {:.3f}s".format(len(r_probe), t1 - t0))

        im = pygdm_visu.vectorfield_color(
            tg.to_np(F.efield)[0], tg.to_np(r_probe), show=0, fieldComp="I"
        )
        plt.colorbar(im)
        plt.show()

        im = pygdm_visu.vectorfield_color(
            tg.to_np(F.get_energy_flux())[0],
            tg.to_np(r_probe),
            show=0,
            fieldComp="I",
        )
        plt.colorbar(im)
        plt.show()

        pygdm_visu.scalarfield(
            [tg.to_np(F.positions)[:, :2], tg.to_np(F.get_chirality())[0]]
        )
        plt.colorbar(im)
        plt.show()
    # %%
    # nf_res = nf(sim, wavelength=wavelength, eval_surface=surface)
    # r_nf = nf_res["tot"]
    # --- farfield

    surface_spherical = tg.tools.geometry.coordinate_map_2d_spherical_lower()
    t0 = time.time()
    ff_res = ff(sim, wavelength=wavelength, r_probe=surface_spherical)
    t1 = time.time()
    print(
        "nf {:}pts, e and h: {:.3f}s".format(
            len(ff_res["e0_config"]) * len(surface_spherical["r_probe"]), t1 - t0
        )
    )

    r_ff = ff_res["sca"]
    I = r_ff.get_efield_intensity()
    I_int = r_ff.get_integrated_efield_intensity()

    # plt.figure(figsize=(15, 2))
    # for i, _I in enumerate(I):
    #     plt.subplot(1, len(I), i + 1)
    #     pygdm_visu.scalarfield([tg.to_np(r_ff.positions)[:, :2], tg.to_np(_I)], show=0)
    # plt.show()

    plt.plot(tg.to_np(I_int))
    plt.show()

    # %%
    delta = 0.1
    field = "e_tot"
    fgrads = field_gradient(sim, wavelength, r_probe, delta=delta, whichfield=field)

    # convert dict to list
    ddx = [fgrads["dfdx"], fgrads["dfdy"], fgrads["dfdz"]]
    field_comps = ["ex", "ey", "ez"]

    # plot field components (real part)
    for i in range(3):
        plt.subplot(3, 4, i * 4 + 1, aspect="equal")
        im = pygdm_visu.vectorfield_color(
            tg.to_np(e_tot)[0],
            tg.to_np(r_probe),
            show=0,
            fieldComp=field_comps[i],
        )
        plt.colorbar(im)

    # plot field gradient components (real part)
    for i in range(3):
        for j in range(3):
            # print(i,j,i * 4 + j + 2)
            plt.subplot(3, 4, i * 4 + j + 2, aspect="equal")
            im = pygdm_visu.vectorfield_color(
                tg.to_np(ddx[j])[0],
                tg.to_np(r_probe),
                show=0,
                fieldComp=field_comps[i],
                cmap="bwr",
            )
            plt.colorbar(im)

    plt.show()
